# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#
# cython: language_level=3
# cython: profile=False
# distutils: language=c

cimport cython
from libc cimport string

cimport acados_solver_common
# TODO: make this import more clear? it is not a general solver, but problem specific.
cimport acados_solver

cimport numpy as cnp

from typing import Tuple, Union, List
import os
import time
from datetime import datetime
import numpy as np


cdef class AcadosOcpSolverCython:
    """
    Class to interact with the acados ocp solver C object.
    """

    cdef acados_solver.nlp_solver_capsule *capsule
    cdef void *nlp_opts
    cdef acados_solver_common.ocp_nlp_dims *nlp_dims
    cdef acados_solver_common.ocp_nlp_config *nlp_config
    cdef acados_solver_common.ocp_nlp_out *nlp_out
    cdef acados_solver_common.ocp_nlp_out *sens_out
    cdef acados_solver_common.ocp_nlp_in *nlp_in
    cdef acados_solver_common.ocp_nlp_solver *nlp_solver

    cdef bint solver_created

    cdef str model_name
    cdef int N
    cdef int status
    cdef double time_solution_sens_lin
    cdef double time_solution_sens_solve
    cdef double time_value_grad

    cdef str nlp_solver_type

    def __cinit__(self, model_name, nlp_solver_type, N):

        self.solver_created = False

        # TODO make N, model_name, status, etc. properties
        self.N = N
        self.model_name = model_name
        self.nlp_solver_type = nlp_solver_type

        # create capsule
        self.capsule = acados_solver.acados_create_capsule()

        # create solver
        assert acados_solver.acados_create(self.capsule) == 0
        self.solver_created = True
        self.status = 0

        # get pointers solver
        self.__get_pointers_solver()


    def __get_pointers_solver(self):
        """
        Private function to get the pointers for solver
        """
        # get pointers solver
        self.nlp_opts = acados_solver.acados_get_nlp_opts(self.capsule)
        self.nlp_dims = acados_solver.acados_get_nlp_dims(self.capsule)
        self.nlp_config = acados_solver.acados_get_nlp_config(self.capsule)
        self.nlp_out = acados_solver.acados_get_nlp_out(self.capsule)
        self.sens_out = acados_solver.acados_get_sens_out(self.capsule)
        self.nlp_in = acados_solver.acados_get_nlp_in(self.capsule)
        self.nlp_solver = acados_solver.acados_get_nlp_solver(self.capsule)



    def solve_for_x0(self, x0_bar, fail_on_nonzero_status=True, print_stats_on_failure=True):
        """
        Wrapper around `solve()` which sets initial state constraint, solves the OCP, and returns u0.
        """
        self.set(0, "lbx", x0_bar)
        self.set(0, "ubx", x0_bar)

        status = self.solve()

        if status != 0:
            if print_stats_on_failure:
                self.print_statistics()
            if fail_on_nonzero_status:
                raise Exception(f'acados acados_ocp_solver returned status {status}')
            elif print_stats_on_failure:
                print(f'Warning: acados acados_ocp_solver returned status {status}')

        u0 = self.get(0, "u")
        return u0


    def solve(self):
        """
        Solve the ocp with current input.
        """
        self.status = acados_solver.acados_solve(self.capsule)
        return self.status


    def get_status(self):
        return self.status

    def reset(self, reset_qp_solver_mem=1):
        """
        Sets current iterate to all zeros.
        """
        return acados_solver.acados_reset(self.capsule, reset_qp_solver_mem)


    def custom_update(self, data_):
        """
        A custom function that can be implemented by a user to be called between solver calls.
        By default this does nothing.
        The idea is to have a convenient wrapper to do complex updates of parameters and numerical data efficiently in C,
        in a function that is compiled into the solver library and can be conveniently used in the Python environment.
        """
        data_len = len(data_)
        cdef cnp.ndarray[cnp.float64_t, ndim=1] data = np.ascontiguousarray(data_, dtype=np.float64)

        return acados_solver.acados_custom_update(self.capsule, <double *> data.data, data_len)


    def set_new_time_steps(self, new_time_steps):
        """
        Set new time steps.
        Recreates the solver if N changes.

            :param new_time_steps: 1 dimensional np array of new time steps for the solver

            .. note:: This allows for different use-cases: either set a new size of time-steps or a new distribution of
                      the shooting nodes without changing the number, e.g., to reach a different final time. Both cases
                      do not require a new code export and compilation.
        """

        raise NotImplementedError("AcadosOcpSolverCython: does not support set_new_time_steps() since it is only a prototyping feature")
        # # unlikely but still possible
        # if not self.solver_created:
        #     raise Exception('Solver was not yet created!')

        # ## check if time steps really changed in value
        # # get time steps
        # cdef cnp.ndarray[cnp.float64_t, ndim=1] old_time_steps = np.ascontiguousarray(np.zeros((self.N,)), dtype=np.float64)
        # assert acados_solver.acados_get_time_steps(self.capsule, self.N, <double *> old_time_steps.data)

        # if np.array_equal(old_time_steps, new_time_steps):
        #     return

        # N = new_time_steps.size
        # cdef cnp.ndarray[cnp.float64_t, ndim=1] value = np.ascontiguousarray(new_time_steps, dtype=np.float64)

        # # check if recreation of acados is necessary (no need to recreate acados if sizes are identical)
        # if len(old_time_steps) == N:
        #     assert acados_solver.acados_update_time_steps(self.capsule, N, <double *> value.data) == 0

        # else:  # recreate the solver with the new time steps
        #     self.solver_created = False

        #     # delete old memory (analog to __del__)
        #     acados_solver.acados_free(self.capsule)

        #     # create solver with new time steps
        #     assert acados_solver.acados_create_with_discretization(self.capsule, N, <double *> value.data) == 0

        #     self.solver_created = True

        #     # get pointers solver
        #     self.__get_pointers_solver()

        # # store time_steps, N
        # self.time_steps = new_time_steps
        # self.N = N


    def update_qp_solver_cond_N(self, qp_solver_cond_N: int):
        """
        Recreate solver with new value `qp_solver_cond_N` with a partial condensing QP solver.
        This function is relevant for code reuse, i.e., if either `set_new_time_steps(...)` is used or
        the influence of a different `qp_solver_cond_N` is studied without code export and compilation.
            :param qp_solver_cond_N: new number of condensing stages for the solver

            .. note:: This function can only be used in combination with a partial condensing QP solver.

            .. note:: After `set_new_time_steps(...)` is used and depending on the new number of time steps it might be
                      necessary to change `qp_solver_cond_N` as well (using this function), i.e., typically
                      `qp_solver_cond_N < N`.
        """
        raise NotImplementedError("AcadosOcpSolverCython: does not support update_qp_solver_cond_N() since it is only a prototyping feature")

        # # unlikely but still possible
        # if not self.solver_created:
        #     raise Exception('Solver was not yet created!')
        # if self.N < qp_solver_cond_N:
        #     raise Exception('Setting qp_solver_cond_N to be larger than N does not work!')
        # if self.qp_solver_cond_N != qp_solver_cond_N:
        #     self.solver_created = False

        #     # recreate the solver
        #     acados_solver.acados_update_qp_solver_cond_N(self.capsule, qp_solver_cond_N)

        #     # store the new value
        #     self.qp_solver_cond_N = qp_solver_cond_N
        #     self.solver_created = True

        #     # get pointers solver
        #     self.__get_pointers_solver()



    def eval_and_get_optimal_value_gradient(self, with_respect_to: str = "initial_state") -> np.ndarray:
        """
        Returns the gradient of the optimal value function w.r.t. what is specified in `with_respect_to`.

        Disclaimer: This function only returns reasonable values if the solver has converged for the current problem instance.

        Notes:
        - for field `initial_state`, the gradient is the Lagrange multiplier of the initial state constraint.
        The gradient computation consists of adding the Lagrange multipliers corresponding to the upper and lower bound of the initial state.

        - for field `params_global`, the gradient of the Lagrange function w.r.t. the global parameters is computed in acados.

        :param with_respect_to: string in ["initial_state", "p_global"]

        """
        if with_respect_to == "params_global":
            print("Deprecation warning: 'params_global' is deprecated and has been renamed to 'p_global'.")
            with_respect_to = "p_global"

        cdef int nx
        cdef int nbu
        cdef int np_global
        cdef int ns_0

        cdef cnp.ndarray[cnp.float64_t, ndim=1] grad
        if with_respect_to == "initial_state":
            # if not self.acados_ocp.constraints.has_x0:
            #     raise Exception("OCP does not have an initial state constraint.")

            nx = acados_solver_common.ocp_nlp_dims_get_from_attr(self.nlp_config, self.nlp_dims, self.nlp_out, 0, "x".encode('utf-8'))
            nbu = acados_solver_common.ocp_nlp_dims_get_from_attr(self.nlp_config, self.nlp_dims, self.nlp_out, 0, "lbu".encode('utf-8'))
            ns_0 = acados_solver_common.ocp_nlp_dims_get_from_attr(self.nlp_config, self.nlp_dims, self.nlp_out, 0, "sl".encode('utf-8'))

            lam = self.get(0, 'lam')
            nlam_non_slack = lam.shape[0]//2 - self.acados_ocp.dims.ns_0
            grad = lam[nbu:nbu+nx] - lam[nlam_non_slack+nbu : nlam_non_slack+nbu+nx]

        elif with_respect_to == "p_global":
            np_global = acados_solver_common.ocp_nlp_dims_get_from_attr(self.nlp_config, self.nlp_dims, self.nlp_out, 0, "p_global".encode('utf-8'))

            field = "p_global".encode('utf-8')
            t0 = time.time()
            grad = np.ascontiguousarray(np.zeros((np_global,)), dtype=np.float64)
            acados_solver_common.ocp_nlp_eval_lagrange_grad_p(self.nlp_solver, self.nlp_in, field, <void *> grad.data)
            self.time_value_grad = time.time() - t0

        else:
            raise Exception(f"AcadosOcpSolver.eval_and_get_optimal_value_gradient(): Unknown field: with_respect_to = {with_respect_to}")
        return grad



    def eval_solution_sensitivity(self, stages: Union[int, List[int]], with_respect_to: str) \
                -> Tuple[Union[List[np.ndarray], np.ndarray], Union[List[np.ndarray], np.ndarray]]:
        """
        Evaluate the sensitivity of the current solution x_i, u_i with respect to the initial state or the parameters for all stages i in `stages`.

            :param stages: stages for which the sensitivities are returned, int or list of int
            :param with_respect_to: string in ["initial_state", "p_global"]
            :returns: a tuple (sens_x, sens_u) with the solution sensitivities.
                    If stages is a list, sens_x is a list of the same length. For sens_u, the list has length len(stages) or len(stages)-1 depending on whether N is included or not.
                    If stages is a scalar, sens_x and sens_u are np.ndarrays of shape (nx[stages], ngrad) and (nu[stages], ngrad).

        .. note::  Correct computation of sensitivities requires \n

        (1) HPIPM as QP solver, \n

        (2) the usage of an exact Hessian, \n

        (3) positive definiteness of the full-space Hessian if the square-root version of the Riccati recursion is used
            OR positive definiteness of the reduced Hessian if the classic Riccati recursion is used (compare: `solver_options.qp_solver_ric_alg`), \n

        (4) the solution of at least one QP in advance to evaluation of the sensitivities as the factorization is reused.

        .. note:: Timing of the sensitivities computation consists of time_solution_sens_lin, time_solution_sens_solve.
        """
        if with_respect_to == "params_global":
            print("Deprecation warning: 'params_global' is deprecated and has been renamed to 'p_global'.")
            with_respect_to = "p_global"

        # if not (self.acados_ocp.solver_options.qp_solver == 'FULL_CONDENSING_HPIPM' or
        #         self.acados_ocp.solver_options.qp_solver == 'PARTIAL_CONDENSING_HPIPM'):
        #     raise Exception("Parametric sensitivities are only available with HPIPM as QP solver.")

        # if not (
        #    (self.acados_ocp.solver_options.hessian_approx == 'EXACT' or
        #    (self.acados_ocp.cost.cost_type == 'LINEAR_LS' and
        #     self.acados_ocp.cost.cost_type_0 == 'LINEAR_LS' and
        #     self.acados_ocp.cost.cost_type_e == 'LINEAR_LS'))
        #     and
        #     self.acados_ocp.solver_options.regularize_method == 'NO_REGULARIZE' and
        #     self.acados_ocp.solver_options.levenberg_marquardt == 0
        # ):
        #     raise Exception("Parametric sensitivities are only correct if an exact Hessian is used!")

        stages_is_list = isinstance(stages, list)
        stages_ = stages if stages_is_list else [stages]

        sens_x = []
        sens_u = []

        cdef int nx
        cdef int nu
        cdef int np_global

        # for s in stages_:
        #     if not isinstance(s, int) or s < 0 or s > N:
        #         raise Exception("AcadosOcpSolver.eval_solution_sensitivity(): stages need to be int or [int] and in [0, N].")

        if with_respect_to == "initial_state":
            nx = acados_solver_common.ocp_nlp_dims_get_from_attr(self.nlp_config, self.nlp_dims, self.nlp_out, 0, "x".encode('utf-8'))

            ngrad = nx
            field = "ex"

        elif with_respect_to == "p_global":
            np_global = acados_solver_common.ocp_nlp_dims_get_from_attr(self.nlp_config, self.nlp_dims, self.nlp_out, 0, "p_global".encode('utf-8'))

            ngrad = np_global
            field = "p_global"

            # compute jacobians wrt params in all modules
            t0 = time.time()
            acados_solver_common.ocp_nlp_eval_params_jac(self.nlp_solver, self.nlp_in, self.nlp_out)
            self.time_solution_sens_lin = time.time() - t0

        else:
            raise Exception(f"AcadosOcpSolver.eval_solution_sensitivity(): Unknown field: with_respect_to = {with_respect_to}")

        # initialize jacobians with zeros
        for s in stages_:
            nx = acados_solver_common.ocp_nlp_dims_get_from_attr(self.nlp_config, self.nlp_dims, self.nlp_out, 0, "x".encode('utf-8'))

            sens_x.append(np.zeros((nx, ngrad)))

            if s < self.N:
                nu = acados_solver_common.ocp_nlp_dims_get_from_attr(self.nlp_config, self.nlp_dims, self.nlp_out, s, "u".encode('utf-8'))
                sens_u.append(np.zeros((nu, ngrad)))


        self.time_solution_sens_solve = 0.0
        for k in range(ngrad):
            # evaluate sensitivity
            acados_solver_common.ocp_nlp_eval_param_sens(self.nlp_solver, field.encode('utf-8'), 0, k, self.sens_out)

            # get timing
            self.time_solution_sens_solve += self.get_stats("time_solution_sensitivities")

            # extract sensitivities
            for n, s in enumerate(stages_):
                sens_x[n][:, k] = self.get(s, "sens_x")

                if s < self.N:
                    sens_u[n][:, k] = self.get(s, "sens_u")

        if not stages_is_list:
            sens_x = sens_x[0]
            sens_u = sens_u[0]

        return sens_x, sens_u



    def eval_param_sens(self, index, stage=0, field="ex"):
        """
        Calculate the sensitivity of the curent solution with respect to the initial state component of index

            :param index: integer corresponding to initial state index in range(nx)
        """

        field_ = field
        field = field_.encode('utf-8')

        # checks
        if not isinstance(index, int):
            raise Exception('AcadosOcpSolverCython.eval_param_sens(): index must be Integer.')

        cdef int nx = acados_solver_common.ocp_nlp_dims_get_from_attr(self.nlp_config, self.nlp_dims, self.nlp_out, 0, "x".encode('utf-8'))

        if index < 0 or index > nx:
            raise Exception(f'AcadosOcpSolverCython.eval_param_sens(): index must be in [0, nx-1], got: {index}.')

        # actual eval_param
        acados_solver_common.ocp_nlp_eval_param_sens(self.nlp_solver, field, stage, index, self.sens_out)

        return


    def get(self, int stage, str field_):
        """
        Get the last solution of the solver:

            :param stage: integer corresponding to shooting node
            :param field: string in ['x', 'u', 'z', 'pi', 'lam', 'sl', 'su',]

            .. note:: regarding lam: \n
                    the inequalities are internally organized in the following order: \n
                    [ lbu lbx lg lh lphi ubu ubx ug uh uphi; \n
                      lsbu lsbx lsg lsh lsphi usbu usbx usg ush usphi]

            .. note:: pi: multipliers for dynamics equality constraints \n
                      lam: multipliers for inequalities \n
                      t: slack variables corresponding to evaluation of all inequalities (at the solution) \n
                      sl: slack variables of soft lower inequality constraints \n
                      su: slack variables of soft upper inequality constraints \n
        """

        out_fields = ['x', 'u', 'z', 'pi', 'lam', 'sl', 'su']
        in_fields = ['p']
        sens_fields = ['sens_u', 'sens_x']
        all_fields = out_fields + in_fields + sens_fields

        if field_ not in all_fields:
            raise Exception(f'AcadosOcpSolver.get(stage={stage}, field={field_}): \'{field_}\' is an invalid argument.\
                    \n Possible values are {all_fields}.')

        if stage < 0 or stage > self.N:
            raise Exception('AcadosOcpSolverCython.get(): stage index must be in [0, N], got: {}.'.format(self.N))

        if stage == self.N and field_ == 'pi':
            raise Exception('AcadosOcpSolverCython.get(): field {} does not exist at final stage {}.'\
                .format(field_, stage))

        field = field_
        if field_ in sens_fields:
            field = field_.replace('sens_', '')
        field = field.encode('utf-8')

        cdef int dims = acados_solver_common.ocp_nlp_dims_get_from_attr(self.nlp_config,
            self.nlp_dims, self.nlp_out, stage, field)

        cdef cnp.ndarray[cnp.float64_t, ndim=1] out = np.zeros((dims,))
        if field_ in out_fields:
            acados_solver_common.ocp_nlp_out_get(self.nlp_config, \
                self.nlp_dims, self.nlp_out, stage, field, <void *> out.data)
        elif field_ in sens_fields:
            acados_solver_common.ocp_nlp_out_get(self.nlp_config, \
                self.nlp_dims, self.sens_out, stage, field, <void *> out.data)
        elif field_ in in_fields:
            acados_solver_common.ocp_nlp_in_get(self.nlp_config, \
                self.nlp_dims, self.nlp_in, stage, field, <void *> out.data)

        return out


    def print_statistics(self):
        """
        prints statistics of previous solver run as a table:
            - iter: iteration number
            - res_stat: stationarity residual
            - res_eq: residual wrt equality constraints (dynamics)
            - res_ineq: residual wrt inequality constraints (constraints)
            - res_comp: residual wrt complementarity conditions
            - qp_stat: status of QP solver
            - qp_iter: number of QP iterations
            - qp_res_stat: stationarity residual of the last QP solution
            - qp_res_eq: residual wrt equality constraints (dynamics) of the last QP solution
            - qp_res_ineq: residual wrt inequality constraints (constraints)  of the last QP solution
            - qp_res_comp: residual wrt complementarity conditions of the last QP solution
        """
        acados_solver.acados_print_stats(self.capsule)


    def store_iterate(self, filename='', overwrite=False, verbose=True):
        """
        Stores the current iterate of the ocp solver in a json file.

            :param filename: if not set, use model_name + timestamp + '.json'
            :param overwrite: if false and filename exists add timestamp to filename
        """
        import json
        if filename == '':
            filename += self.model_name + '_' + 'iterate' + '.json'

        if not overwrite:
            # append timestamp
            if os.path.isfile(filename):
                filename = filename[:-5]
                filename += datetime.now().strftime('%Y-%m-%d-%H:%M:%S.%f') + '.json'

        # get iterate:
        solution = dict()

        lN = len(str(self.N+1))
        for i in range(self.N+1):
            i_string = f'{i:0{lN}d}'
            solution['x_'+i_string] = self.get(i,'x')
            solution['u_'+i_string] = self.get(i,'u')
            solution['z_'+i_string] = self.get(i,'z')
            solution['lam_'+i_string] = self.get(i,'lam')
            solution['sl_'+i_string] = self.get(i, 'sl')
            solution['su_'+i_string] = self.get(i, 'su')
            if i < self.N:
                solution['pi_'+i_string] = self.get(i,'pi')

        for k in list(solution.keys()):
            if len(solution[k]) == 0:
                del solution[k]

        # save
        with open(filename, 'w') as f:
            json.dump(solution, f, default=lambda x: x.tolist(), indent=4, sort_keys=True)
        if verbose:
            print("stored current iterate in ", os.path.join(os.getcwd(), filename))


    def load_iterate(self, filename, verbose:bool=False):
        """
        Loads the iterate stored in json file with filename into the ocp solver.
        """
        import json
        if not os.path.isfile(filename):
            raise Exception('load_iterate: failed, file does not exist: ' + os.path.join(os.getcwd(), filename))

        with open(filename, 'r') as f:
            solution = json.load(f)

        if verbose:
            print(f"loading iterate {filename}")

        for key in solution.keys():
            (field, stage) = key.split('_')
            self.set(int(stage), field, np.array(solution[key]))


    def get_stats(self, field_):
        """
        Get the information of the last solver call.

            :param field: string in ['statistics', 'time_tot', 'time_lin', 'time_sim', 'time_sim_ad', 'time_sim_la', 'time_qp', 'time_qp_solver_call', 'time_reg', 'sqp_iter']
        Available fileds:
            - time_tot: total CPU time previous call
            - time_lin: CPU time for linearization
            - time_sim: CPU time for integrator
            - time_sim_ad: CPU time for integrator contribution of external function calls
            - time_sim_la: CPU time for integrator contribution of linear algebra
            - time_qp: CPU time qp solution
            - time_qp_solver_call: CPU time inside qp solver (without converting the QP)
            - time_qp_xcond: time_glob: CPU time globalization
            - time_solution_sensitivities: CPU time for previous call to eval_param_sens
            - time_reg: CPU time regularization
            - sqp_iter: number of SQP iterations
            - qp_iter: vector of QP iterations for last SQP call
            - statistics: table with info about last iteration
            - stat_m: number of rows in statistics matrix
            - stat_n: number of columns in statistics matrix
            - residuals: residuals of last iterate
            - alpha: step sizes of SQP iterations
        """

        double_fields = ['time_tot',
                  'time_lin',
                  'time_sim',
                  'time_sim_ad',
                  'time_sim_la',
                  'time_qp',
                  'time_qp_solver_call',
                  'time_qp_xcond',
                  'time_glob',
                  'time_solution_sensitivities',
                  'time_reg'
        ]
        fields = double_fields + [
                  'sqp_iter',
                  'qp_iter',
                  'statistics',
                  'stat_m',
                  'stat_n',
                  'residuals',
                  'alpha',
                ]
        field = field_.encode('utf-8')

        if field_ in ['sqp_iter', 'stat_m', 'stat_n']:
            return self.__get_stat_int(field)

        elif field_ in double_fields:
            return self.__get_stat_double(field)

        elif field_ == 'statistics':
            sqp_iter = self.get_stats("sqp_iter")
            stat_m = self.get_stats("stat_m")
            stat_n = self.get_stats("stat_n")
            min_size = min([stat_m, sqp_iter+1])
            return self.__get_stat_matrix(field, stat_n+1, min_size)

        elif field_ == 'qp_iter':
            full_stats = self.get_stats('statistics')
            if self.nlp_solver_type == 'SQP':
                return full_stats[6, :]
            elif self.nlp_solver_type == 'SQP_RTI':
                return full_stats[2, :]

        elif field_ == 'alpha':
            full_stats = self.get_stats('statistics')
            if self.nlp_solver_type == 'SQP':
                return full_stats[7, :]
            else: # self.nlp_solver_type == 'SQP_RTI':
                raise Exception("alpha values are not available for SQP_RTI")

        elif field_ == 'residuals':
            return self.get_residuals()

        else:
            raise NotImplementedError("TODO!")


    def __get_stat_int(self, field):
        cdef int out
        acados_solver_common.ocp_nlp_get(self.nlp_solver, field, <void *> &out)
        return out

    def __get_stat_double(self, field):
        cdef double out
        acados_solver_common.ocp_nlp_get(self.nlp_solver, field, <void *> &out)
        return out

    def __get_stat_matrix(self, field, n, m):
        cdef cnp.ndarray[cnp.float64_t, ndim=2] out_mat = np.ascontiguousarray(np.zeros((n, m)), dtype=np.float64)
        acados_solver_common.ocp_nlp_get(self.nlp_solver, field, <void *> out_mat.data)
        return out_mat


    def get_cost(self):
        """
        Returns the cost value of the current solution.
        """
        # compute cost internally
        acados_solver_common.ocp_nlp_eval_cost(self.nlp_solver, self.nlp_in, self.nlp_out)

        # create output
        cdef double out

        # call getter
        acados_solver_common.ocp_nlp_get(self.nlp_solver, "cost_value", <void *> &out)

        return out


    def get_residuals(self, recompute=False):
        """
        Returns an array of the form [res_stat, res_eq, res_ineq, res_comp].
        """
        # compute residuals if RTI
        if self.nlp_solver_type == 'SQP_RTI' or recompute:
            acados_solver_common.ocp_nlp_eval_residuals(self.nlp_solver, self.nlp_in, self.nlp_out)

        # create output array
        cdef cnp.ndarray[cnp.float64_t, ndim=1] out = np.ascontiguousarray(np.zeros((4,), dtype=np.float64))
        cdef double double_value

        field = "res_stat".encode('utf-8')
        acados_solver_common.ocp_nlp_get(self.nlp_solver, field, <void *> &double_value)
        out[0] = double_value

        field = "res_eq".encode('utf-8')
        acados_solver_common.ocp_nlp_get(self.nlp_solver, field, <void *> &double_value)
        out[1] = double_value

        field = "res_ineq".encode('utf-8')
        acados_solver_common.ocp_nlp_get(self.nlp_solver, field, <void *> &double_value)
        out[2] = double_value

        field = "res_comp".encode('utf-8')
        acados_solver_common.ocp_nlp_get(self.nlp_solver, field, <void *> &double_value)
        out[3] = double_value

        return out


    # Note: this function should not be used anymore, better use cost_set, constraints_set
    def set(self, int stage, str field_, value_):
        """
        Set numerical data inside the solver.

            :param stage: integer corresponding to shooting node
            :param field: string in ['x', 'u', 'pi', 'lam', 'p']

            .. note:: regarding lam: \n
                    the inequalities are internally organized in the following order: \n
                    [ lbu lbx lg lh lphi ubu ubx ug uh uphi; \n
                      lsbu lsbx lsg lsh lsphi usbu usbx usg ush usphi]

            .. note:: pi: multipliers for dynamics equality constraints \n
                      lam: multipliers for inequalities \n
                      t: slack variables corresponding to evaluation of all inequalities (at the solution) \n
                      sl: slack variables of soft lower inequality constraints \n
                      su: slack variables of soft upper inequality constraints \n
        """
        if not isinstance(value_, np.ndarray):
            raise Exception(f"set: value must be numpy array, got {type(value_)}.")
        cost_fields = ['y_ref', 'yref']
        constraints_fields = ['lbx', 'ubx', 'lbu', 'ubu']
        out_fields = ['x', 'u', 'pi', 'lam', 'z', 'sl', 'su']
        mem_fields = ['xdot_guess', 'z_guess']

        field = field_.encode('utf-8')

        cdef cnp.ndarray[cnp.float64_t, ndim=1] value = np.ascontiguousarray(value_, dtype=np.float64)

        # treat parameters separately
        if field_ == 'p':
            assert acados_solver.acados_update_params(self.capsule, stage, <double *> value.data, value.shape[0]) == 0
        else:
            if field_ not in constraints_fields + cost_fields + out_fields:
                raise Exception("AcadosOcpSolverCython.set(): {} is not a valid argument.\
                    \nPossible values are {}.".format(field, \
                    constraints_fields + cost_fields + out_fields + ['p']))

            dims = acados_solver_common.ocp_nlp_dims_get_from_attr(self.nlp_config,
                self.nlp_dims, self.nlp_out, stage, field)

            if value_.shape[0] != dims:
                msg = 'AcadosOcpSolverCython.set(): mismatching dimension for field "{}" '.format(field_)
                msg += 'with dimension {} (you have {})'.format(dims, value_.shape[0])
                raise Exception(msg)

            if field_ in constraints_fields:
                acados_solver_common.ocp_nlp_constraints_model_set(self.nlp_config,
                    self.nlp_dims, self.nlp_in, stage, field, <void *> value.data)
            elif field_ in cost_fields:
                acados_solver_common.ocp_nlp_cost_model_set(self.nlp_config,
                    self.nlp_dims, self.nlp_in, stage, field, <void *> value.data)
            elif field_ in out_fields:
                acados_solver_common.ocp_nlp_out_set(self.nlp_config,
                    self.nlp_dims, self.nlp_out, stage, field, <void *> value.data)
            elif field_ in mem_fields:
                acados_solver_common.ocp_nlp_set(self.nlp_solver, stage, field, <void *> value.data)

            if field_ == 'z':
                field = 'z_guess'.encode('utf-8')
                acados_solver_common.ocp_nlp_set(self.nlp_solver, stage, field, <void *> value.data)
        return

    def cost_set(self, int stage, str field_, value_):
        """
        Set numerical data in the cost module of the solver.

            :param stage: integer corresponding to shooting node
            :param field: string, e.g. 'yref', 'W', 'ext_cost_num_hess', 'scaling'
            :param value: of appropriate size
        """
        if not isinstance(value_, np.ndarray):
            raise Exception(f"cost_set: value must be numpy array, got {type(value_)}.")
        field = field_.encode('utf-8')

        cdef int dims[2]
        acados_solver_common.ocp_nlp_cost_dims_get_from_attr(self.nlp_config, \
            self.nlp_dims, self.nlp_out, stage, field, &dims[0])

        cdef double[::1,:] value

        value_shape = value_.shape
        if len(value_shape) == 1:
            value_shape = (value_shape[0], 0)
            value = np.asfortranarray(value_[None,:])

        elif len(value_shape) == 2:
            # Get elements in column major order
            value = np.asfortranarray(value_)

        if value_shape[0] != dims[0] or value_shape[1] != dims[1]:
            raise Exception('AcadosOcpSolverCython.cost_set(): mismatching dimension' +
                f' for field "{field_}" at stage {stage} with dimension {tuple(dims)} (you have {value_shape})')

        acados_solver_common.ocp_nlp_cost_model_set(self.nlp_config, \
            self.nlp_dims, self.nlp_in, stage, field, <void *> &value[0][0])


    def constraints_set(self, int stage, str field_, value_):
        """
        Set numerical data in the constraint module of the solver.

            :param stage: integer corresponding to shooting node
            :param field: string in ['lbx', 'ubx', 'lbu', 'ubu', 'lg', 'ug', 'lh', 'uh', 'uphi', 'C', 'D']
            :param value: of appropriate size
        """
        if not isinstance(value_, np.ndarray):
            raise Exception(f"constraints_set: value must be numpy array, got {type(value_)}.")

        field = field_.encode('utf-8')

        cdef int dims[2]
        acados_solver_common.ocp_nlp_constraint_dims_get_from_attr(self.nlp_config, \
            self.nlp_dims, self.nlp_out, stage, field, &dims[0])

        cdef double[::1,:] value

        value_shape = value_.shape
        if len(value_shape) == 1:
            value_shape = (value_shape[0], 0)
            value = np.asfortranarray(value_[None,:])

        elif len(value_shape) == 2:
            # Get elements in column major order
            value = np.asfortranarray(value_)

        if value_shape != tuple(dims):
            raise Exception(f'AcadosOcpSolverCython.constraints_set(): mismatching dimension' +
                f' for field "{field_}" at stage {stage} with dimension {tuple(dims)} (you have {value_shape})')

        acados_solver_common.ocp_nlp_constraints_model_set(self.nlp_config, \
            self.nlp_dims, self.nlp_in, stage, field, <void *> &value[0][0])

        return


    def get_from_qp_in(self, int stage, str field_):
        """
        Get numerical data from the dynamics module of the solver:

            :param stage: integer corresponding to shooting node
            :param field: string, e.g. 'A'
        """
        field = field_.encode('utf-8')

        # get dims
        cdef int[2] dims
        acados_solver_common.ocp_nlp_qp_dims_get_from_attr(self.nlp_config, self.nlp_dims, self.nlp_out, stage, field, &dims[0])

        # create output data
        cdef cnp.ndarray[cnp.float64_t, ndim=2] out = np.zeros((dims[0], dims[1]), order='F')

        # call getter
        acados_solver_common.ocp_nlp_get_at_stage(self.nlp_solver, stage, field, <void *> out.data)

        return out


    def options_set(self, str field_, value_):
        """
        Set options of the solver.

            :param field: string, e.g. 'print_level', 'rti_phase', 'step_length', 'alpha_min', 'alpha_reduction', 'qp_warm_start', 'line_search_use_sufficient_descent', 'full_step_dual', 'globalization_use_SOC', 'qp_tol_stat', 'qp_tol_eq', 'qp_tol_ineq', 'qp_tol_comp', 'qp_tau_min', 'qp_mu0'

            :param value: of type int, float, string

            - qp_tol_stat: QP solver tolerance stationarity
            - qp_tol_eq: QP solver tolerance equalities
            - qp_tol_ineq: QP solver tolerance inequalities
            - qp_tol_comp: QP solver tolerance complementarity
            - qp_tau_min: for HPIPM QP solvers: minimum value of barrier parameter in HPIPM
            - qp_mu0: for HPIPM QP solvers: initial value for complementarity slackness
            - warm_start_first_qp: indicates if first QP in SQP is warm_started
        """
        int_fields = ['print_level', 'rti_phase', 'qp_warm_start', 'line_search_use_sufficient_descent', 'full_step_dual', 'globalization_use_SOC', 'warm_start_first_qp']
        double_fields = ['step_length', 'tol_eq', 'tol_stat', 'tol_ineq', 'tol_comp', 'alpha_min', 'alpha_reduction', 'eps_sufficient_descent',
        'qp_tol_stat', 'qp_tol_eq', 'qp_tol_ineq', 'qp_tol_comp', 'qp_tau_min', 'qp_mu0']
        string_fields = ['globalization']

        # encode
        field = field_.encode('utf-8')

        cdef int int_value
        cdef double double_value
        cdef unsigned char[::1] string_value

        # check field availability and type
        if field_ in int_fields:
            if not isinstance(value_, int):
                raise Exception('solver option {} must be of type int. You have {}.'.format(field_, type(value_)))

            if field_ == 'rti_phase':
                if value_ < 0 or value_ > 2:
                    raise Exception('AcadosOcpSolverCython.solve(): argument \'rti_phase\' can '
                        'take only values 0, 1, 2 for SQP-RTI-type solvers')
                if self.nlp_solver_type != 'SQP_RTI':
                    raise Exception('AcadosOcpSolverCython.solve(): argument \'rti_phase\' can '
                        'take only value 0 for SQP-type solvers')

            int_value = value_
            acados_solver_common.ocp_nlp_solver_opts_set(self.nlp_config, self.nlp_opts, field, <void *> &int_value)

        elif field_ in double_fields:
            if not isinstance(value_, float):
                raise Exception('solver option {} must be of type float. You have {}.'.format(field_, type(value_)))

            double_value = value_
            acados_solver_common.ocp_nlp_solver_opts_set(self.nlp_config, self.nlp_opts, field, <void *> &double_value)

        elif field_ in string_fields:
            if not isinstance(value_, bytes):
                raise Exception('solver option {} must be of type str. You have {}.'.format(field_, type(value_)))

            string_value = value_.encode('utf-8')
            acados_solver_common.ocp_nlp_solver_opts_set(self.nlp_config, self.nlp_opts, field, <void *> &string_value[0])

        else:
            raise Exception('AcadosOcpSolverCython.options_set() does not support field {}.'\
                '\n Possible values are {}.'.format(field_, ', '.join(int_fields + double_fields + string_fields)))


    def set_params_sparse(self, int stage, idx_values_, param_values_):
        """
        set parameters of the solvers external function partially:
        Pseudo: solver.param[idx_values_] = param_values_;
        Parameters:
            :param stage_: integer corresponding to shooting node
            :param idx_values_: 0 based integer array corresponding to parameter indices to be set
            :param param_values_: new parameter values as numpy array
        """

        if not isinstance(param_values_, np.ndarray):
            raise Exception('param_values_ must be np.array.')

        if param_values_.shape[0] != len(idx_values_):
            raise Exception(f'param_values_ and idx_values_ must be of the same size.' +
                 f' Got sizes idx {param_values_.shape[0]}, param_values {len(idx_values_)}.')

        # n_update = c_int(len(param_values_))

        # param_data = cast(param_values_.ctypes.data, POINTER(c_double))
        # c_idx_values = np.ascontiguousarray(idx_values_, dtype=np.intc)
        # idx_data = cast(c_idx_values.ctypes.data, POINTER(c_int))

        # getattr(self.shared_lib, f"{self.model_name}_acados_update_params_sparse").argtypes = \
        #                 [c_void_p, c_int, POINTER(c_int), POINTER(c_double), c_int]
        # getattr(self.shared_lib, f"{self.model_name}_acados_update_params_sparse").restype = c_int
        # getattr(self.shared_lib, f"{self.model_name}_acados_update_params_sparse") \
        #                             (self.capsule, stage, idx_data, param_data, n_update)

        cdef cnp.ndarray[cnp.float64_t, ndim=1] value = np.ascontiguousarray(param_values_, dtype=np.float64)
        # cdef cnp.ndarray[cnp.intc, ndim=1] idx = np.ascontiguousarray(idx_values_, dtype=np.intc)

        # NOTE: this does throw an error somehow:
        # ValueError: Buffer dtype mismatch, expected 'int object' but got 'int'
        # cdef cnp.ndarray[cnp.int, ndim=1] idx = np.ascontiguousarray(idx_values_, dtype=np.intc)

        cdef cnp.ndarray[cnp.int32_t, ndim=1] idx = np.ascontiguousarray(idx_values_, dtype=np.int32)
        cdef int n_update = value.shape[0]
        # print(f"in set_params_sparse Cython n_update {n_update}")

        assert acados_solver.acados_update_params_sparse(self.capsule, stage, <int *> idx.data, <double *> value.data, n_update) == 0
        return


    def set_p_global_and_precompute_dependencies(self, data_):
        """
        Sets values of p_global and precomputes all parts of the CasADi graphs of all other functions that only depend on p_global.
        """
        cdef cnp.ndarray[cnp.float64_t, ndim=1] value = np.ascontiguousarray(data_, dtype=np.float64)

        cdef int data_len = value.shape[0]

        cdef int status = acados_solver.acados_set_p_global_and_precompute_dependencies(self.capsule, <double *> value.data, data_len)

        return status


    def __del__(self):
        if self.solver_created:
            acados_solver.acados_free(self.capsule)
            acados_solver.acados_free_capsule(self.capsule)



#
#     This file is part of rockit.
#
#     rockit -- Rapid Optimal Control Kit
#     Copyright (C) 2019 MECO, KU Leuven. All rights reserved.
#
#     Rockit is free software; you can redistribute it and/or
#     modify it under the terms of the GNU Lesser General Public
#     License as published by the Free Software Foundation; either
#     version 3 of the License, or (at your option) any later version.
#
#     Rockit is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#     Lesser General Public License for more details.
#
#     You should have received a copy of the GNU Lesser General Public
#     License along with CasADi; if not, write to the Free Software
#     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
#

from ...freetime import FreeTime
from ...casadi_helpers import prepare_build_dir
from ..method import ExternalMethod, linear_coeffs, check_Js
from ...solution import OcpSolution

import numpy as np
from casadi import external, vec, CodeGenerator, SX, Sparsity, MX, vcat, veccat, symvar, substitute, densify, sparsify, DM, Opti, is_linear, vertcat, depends_on, jacobian, linear_coeff, quadratic_coeff, mtimes, pinv, evalf, Function, vvcat, inf, sum1, sum2, diag, solve, fmin, fmax
import casadi
import casadi as cs
from ...casadi_helpers import DM2numpy, reshape_number
from collections import OrderedDict

import subprocess
import os
from ctypes import *
import glob
import shutil
import platform
import hashlib

def get_terms(e):
    def get_terms_internal(e):
        if e.op()==cs.OP_ADD:
            for i in range(e.n_dep()):
                for t in get_terms_internal(e.dep(i)):
                    yield t
        else:
            yield e
    return list(get_terms_internal(e))


def visit(e,parents=None):
    if parents is None:
        parents = []
    yield (e,parents)
    for i in range(e.n_dep()):
        for t in visit(e.dep(i),parents=[e]+parents):
            yield t

"""
conda install -c conda-forge make cmake
"""

external_dir = os.path.join(os.path.dirname(__file__),"external")

if not os.path.exists(external_dir):
    raise Exception("GRAMPC source not found")

windows_generator = None
if platform.system()=="Windows":
    if "VCVARS" in os.environ:
        vcvars = os.environ["VCVARS"]
    else:
        vcvars = None
        def vcvars_dirs():
            for p in ["Program Files (x86)","Program Files"]:
                for suffix in ["","Community\\","Professional\\"]:
                    yield f"C:\\{p}\\Microsoft Visual Studio\\2022\\{suffix}VC\Auxiliary\\Build\\vcvars64.bat"
                    yield f"C:\\{p}\\Microsoft Visual Studio\\2019\\{suffix}VC\Auxiliary\\Build\\vcvars64.bat"
                    yield f"C:\\{p}\\Microsoft Visual Studio\\2017\\{suffix}VC\Auxiliary\\Build\\vcvars64.bat"
        for e in vcvars_dirs():
            if os.path.exists(e):
                vcvars = e
                break
        if vcvars is None:
            raise Exception("Could not find vcvars bat file. Tried %s." % str(list(vcvars_dirs())))

        if "2022" in vcvars:
            windows_generator = "Visual Studio 17 2022"
        if "2019" in vcvars:
            windows_generator = "Visual Studio 16 2019"
        if "2017" in vcvars:
            windows_generator = "Visual Studio 15 2017"
    cmd = f'"{vcvars}" && "build.bat"'

def run_build(cwd=None):
    if platform.system()=="Windows":
        cmd = f'"{vcvars}" && "build.bat"'
        print(cmd)
        subprocess.run(cmd,shell=True,cwd=cwd)
    else:
        subprocess.run(["bash","build.sh"],cwd=cwd)

INF = 1e20
"""
In general, the constraints should be formulated in such a way that there are no conflicts. However,
numerical difficulties can arise in some problems if constraints are formulated twice for the last point.
Therefore, GRAMPC does not evaluate the constraints g and h for the last trajectory point if terminal
constraints are defined, i.e. NgT + NhT > 0. In contrast, if no terminal constraints are defined, the
functions g and h are evaluated for all points. Note that the opposite behavior is easy to implement
by including g and h in the terminal constraints gT and hT .


TODO: debug with example from repo, make callbacks print

The convergence checks are a bit wird. The gradient of the Lagrangian is not checked; a small step size in decision space is enough

GRAMPC is an indirect approach (first optimize then discretize).
The boundary value problem is discretised using single shooting (we integrate over the entire time horizon).
The inner part of the min-max problem is solved using a projected gradient method (first order).lsAdapt


Debugging test1.py
Suspected that convergence ends prematurely due to poorly implemented linesearch.
Played around with linesearch options, also experimented with resetting the adaptive strategy after alpha shrinks a lot -> no difference

"""


stats_fields = [('stop_crit',c_int),('conv_grad',c_int),('conv_con',c_int),('n_outer_iter',c_int),('n_inner_iter',c_int),('runtime',c_double)]

class StatsStruct(Structure):
    _fields_ = stats_fields


def format_float(e):
    return "%0.18f" % e

def strlist(a):
    elems = []
    for e in a:
        if isinstance(e,str):
            elems.append('"'+e+'"')
        elif isinstance(e,float):
            elems.append(format_float(e))
        else:
            elems.append(str(e))
    return ",".join(elems)
    
def check_Js(J):
    """
    Checks if J, a pre-multiplier for slacks, is of legitimate structure
    Empty rows are allowed
    """
    try:
        J = evalf(J)
    except:
        raise Exception("Slack error")
    assert np.all(np.array(J.nonzeros())==1), "All nonzeros must be 1"
    # Check if slice of permutation of unit matrix
    assert np.all(np.array(sum2(J))<=1), "Each constraint can only depend on one slack at most"
    assert np.all(np.array(sum1(J))<=1), "Each constraint must depend on a unique slack, if any"


def mark(slack, Js):
    assert np.all(np.array(slack[Js.sparsity().get_col()])==0)
    slack[Js.sparsity().get_col()] = 1

def export_expr(m):
    if isinstance(m,list):
        if len(m)==0:
            return MX(0, 1)
        else:
            return vcat(m)
    return m

def export_num(m):
    res=np.array(evalf(export_expr(m)))
    if np.any(res==-inf) or np.any(res==inf):
        print("WARNING: Double-sided constraints are much preferred. Replaced inf with %f." % INF)
    res[res==-inf] = -INF
    res[res==inf] = INF
    return res

def export_num_vec(m):
    return np.array(evalf(export_expr(m))).reshape(-1)

def export(m):
    return (export_expr(m),False)

def export_vec(m):
    return (export_expr(m),True)

class MyCodeGenerator:
    def __init__(self,name):
        self.added_shorthands = set()
        self.add_includes = []
        self.prefix = ""
        self.auxiliaries = ""
        self.body = ""
        self.name = name

    def add_include(self,h):
        self.add_includes.append(h)

    def shorthand(self,name):
        self.added_shorthands.add(name)
        return self.prefix + name

    def add_dependency(self, f):
        name = f.codegen_name(self, False)
        fname = self.prefix + name
        stack_counter = self.shorthand(name + "_unused_stack_counter")
        stack = self.shorthand(name + "_unused_stack")
        mem_counter = self.shorthand(name + "_mem_counter")
        mem_array = self.shorthand(name + "_mem")
        alloc_mem = self.shorthand(name + "_alloc_mem")
        init_mem = self.shorthand(name + "_init_mem")
        work = self.shorthand(name+"_work")

        self.auxiliaries += f"static int {mem_counter} = 0;\n"
        self.auxiliaries += f"static int {stack_counter } = -1;\n"
        self.auxiliaries += f"static int {stack}[CASADI_MAX_NUM_THREADS];\n"
        self.auxiliaries += f"static {f.codegen_mem_type()} *{mem_array}[CASADI_MAX_NUM_THREADS];\n\n"

        f.codegen_declarations(self)

        f.codegen(self, fname)

        def encode_sp(sp,i):
            r = f"case {i}:\n"
            r+= "{"
            spc = sp.compress()
            r+= f"static casadi_int sp[{len(spc)}] = {{{strlist(spc)}}}; return sp;"
            r+= "}"
            return r

        def encode_name(n,i):
            return f"case {i}: return \"{n}\";\n"

        newline = "\n"

        self.body += f"""

            casadi_int {fname}_n_in(void) {{
                return {f.n_in()};
            }}

            casadi_int {fname}_n_out(void) {{
                return {f.n_out()};
            }}

            const casadi_int* {fname}_sparsity_in(casadi_int i) {{
                switch (i) {{
                {newline.join(encode_sp(f.sparsity_in(i), i) for i in range(f.n_in()))}
                default: return 0;
                }}
            }}

            const casadi_int* {fname}_sparsity_out(casadi_int i) {{
                switch (i) {{
                {newline.join(encode_sp(f.sparsity_out(i), i) for i in range(f.n_out()))}
                default: return 0;
                }}
            }}

            const char* {fname}_name_in(casadi_int i) {{
                switch (i) {{
                {newline.join(encode_name(f.name_in(i), i) for i in range(f.n_in()))}
                default: return 0;
                }}
            }}

            const char* {fname}_name_out(casadi_int i) {{
                switch (i) {{
                {newline.join(encode_name(f.name_out(i), i) for i in range(f.n_out()))}
                default: return 0;
                }}
            }}

            int {fname}_work(casadi_int *sz_arg, casadi_int* sz_res, casadi_int *sz_iw, casadi_int *sz_w) {{
            {f.codegen_work(self)}
              return 0;
            }}

            // Alloc memory
            int {fname}_alloc_mem(void) {{
            {f.codegen_alloc_mem(self)}
            }}

            // Initialize memory
            int {fname}_init_mem(int mem) {{
            {f.codegen_init_mem(self)}
            }}

            // Clear memory
            void {fname}_free_mem(int mem) {{
            {f.codegen_free_mem(self)}
            }}

            int {self.shorthand(name + "_checkout")}(void) {{
            int mid;
            if ({stack_counter}>=0) {{
              return {stack}[{stack_counter}--];
            }} else {{
              if ({mem_counter}==CASADI_MAX_NUM_THREADS) return -1;
              mid = {alloc_mem}();
              if (mid<0) return -1;
              if({init_mem}(mid)) return -1;
              return mid;
            }}

            return {stack}[{stack_counter}--];
        }}

        void {self.shorthand(name+"_release")}(int mem) {{
            {stack}[++{stack_counter}] = mem;
        }}


        """

    def generate(self,dir="."):
        with open(os.path.join(dir,self.name+".c"),"w") as out:
            out.write("#define CASADI_MAX_NUM_THREADS 1\n")
            for e in self.add_includes:
                out.write(f"#include \"{e}\"\n")
            out.write(self.auxiliaries)
            out.write(self.body)

class Wrapper:
    def __init__(self,userparam):
        self.added_declarations = ""
        self.added_init_mem = ""
        self.added_body = ""
        self.sp_in = []
        self.sp_out = []
        self._name_in = []
        self._name_out = []
        self.userparam = userparam

    def set_sp_in(self, sp):
        self.sp_in = sp

    def set_sp_out(self, sp):
        self.sp_out = sp

    def set_name_in(self, name):
        self._name_in = name

    def set_name_out(self, name):
        self._name_out = name

    def add_declarations(self, decl):
        self.added_declarations += decl
    
    def add_init_mem(self, init_mem):
        self.added_init_mem += init_mem

    def add_body(self, body):
        self.added_body += body

    def signature(self, fname):
        return "int " + fname + "(const casadi_real** arg, casadi_real** res, casadi_int* iw, casadi_real* w, int mem)"

    def codegen(self, g, fname):
        g.body += f"{self.signature(fname)} {{\n"
        self.codegen_body(g)
        g.body += f"return 0;}}\n"

    def codegen_name(self, g, ns):
        return "grampc_driver"
    def codegen_mem_type(self):
        return "typeGRAMPC"
    def codegen_alloc_mem(self, g):
        name = self.codegen_name(g, False)
        mem_counter = g.shorthand(name + "_mem_counter")
        return f"return {mem_counter}++;"

    def codegen_init_mem(self, g):
        grampc = self.codegen_mem(g, "mem")
        userparam = grampc+"->userparam"
        return f"""
        typeGRAMPC *grampc = {grampc};
        {self.added_init_mem}
        {grampc} = grampc;
        return 0;
        
        """

    def codegen_free_mem(self, g):
        grampc = self.codegen_mem(g)
        userparam = grampc+"->userparam"
        return f"postamble({userparam});\nfree({userparam});"

    def codegen_mem(self, g, index=0):
        name = self.codegen_name(g, False)
        mem_array = g.shorthand(name + "_mem")
        return mem_array+"[" + str(index) + "]"

    def codegen_body(self, g):
        grampc = self.codegen_mem(g, "mem")
        g.add_include("grampc.h")
        g.body += f"typeGRAMPC *grampc = {grampc};"
        stats = f"{self.userparam}->stats"
        g.body += f"{stats}.stop_crit=0;"
        g.body += f"{stats}.n_outer_iter=0;"
        g.body += f"{stats}.n_inner_iter=0;"
        g.body += f"{stats}.runtime=0;"
        g.body += self.added_body

    def n_in(self):
        return len(self.sp_in)

    def n_out(self):
        return len(self.sp_out)

    def sparsity_in(self, i):
        return self.sp_in[i]

    def sparsity_out(self, i):
        return self.sp_out[i]

    def name_in(self, i):
        return self._name_in[i]

    def name_out(self, i):
        return self._name_out[i]


    def codegen_work(self, g):
        r = f"""
            casadi_int sz_arg_local, sz_res_local, sz_iw_local, sz_w_local;
            *sz_arg=0, *sz_res=0, *sz_iw=0, *sz_w=0;
            pmap_work(&sz_arg_local, &sz_res_local, &sz_iw_local, &sz_w_local);
            if (sz_arg_local>*sz_arg) *sz_arg=sz_arg_local;
            if (sz_res_local>*sz_res) *sz_res=sz_res_local;
            if (sz_iw_local>*sz_iw) *sz_iw=sz_iw_local;
            if (sz_w_local>*sz_iw) *sz_w=sz_w_local;

            *sz_arg += {self.n_in()};
            *sz_res += {self.n_out()};
        """
        return r


    def codegen_declarations(self, g):
        g.auxiliaries += """


        typedef struct {
            int stop_crit;
            int conv_grad;
            int conv_con;
            int n_outer_iter;
            int n_inner_iter;
            double runtime;
        } grampc_solver_stats;

        typedef struct {
            int sqp_stop_crit;
            int n_sqp_iter;
            int n_ls;
            int n_max_ls;
            int n_qp_iter;
            int n_max_qp_iter;
            double runtime;
        } compat_solver_stats;

        typedef struct cs_struct_def {
            const casadi_real** arg;
            casadi_real** res;
            casadi_int* iw;
            casadi_real* w;
            casadi_real* p;
            typeGRAMPC* grampc;
            casadi_real* x_opt;
            casadi_real* u_opt;
            casadi_real T_opt;
            casadi_real* v_opt;
            casadi_real* x_current;
            casadi_real* umin;
            casadi_real* umax;
            casadi_real* u0;
            casadi_real* v0;
            casadi_real Tmin;
            casadi_real Tmax;
            grampc_solver_stats stats;
        } cs_struct;

        /** Additional functions required for semi-implicit systems 
            M*dx/dt(t) = f(t0+t,x(t),u(t),p) using the solver RODAS 
            ------------------------------------------------------- **/
        /** Jacobian df/dx in vector form (column-wise) **/
        void dfdx(typeRNum *out, ctypeRNum t, ctypeRNum *x, ctypeRNum *u, ctypeRNum *p, typeUSERPARAM *userparam)
        {
        }
        /** Jacobian df/dx in vector form (column-wise) **/
        void dfdxtrans(typeRNum *out, ctypeRNum t, ctypeRNum *x, ctypeRNum *u, ctypeRNum *p, typeUSERPARAM *userparam)
        {
        }
        /** Jacobian df/dt **/
        void dfdt(typeRNum *out, ctypeRNum t, ctypeRNum *x, ctypeRNum *u, ctypeRNum *p, typeUSERPARAM *userparam)
        {
        }
        /** Jacobian d(dH/dx)/dt  **/
        void dHdxdt(typeRNum *out, ctypeRNum t, ctypeRNum *x, ctypeRNum *u, ctypeRNum *vec, ctypeRNum *p, typeUSERPARAM *userparam)
        {
        }
        /** Mass matrix in vector form (column-wise, either banded or full matrix) **/
        void Mfct(typeRNum *out, typeUSERPARAM *userparam)
        {
        }
        /** Transposed mass matrix in vector form (column-wise, either banded or full matrix) **/
        void Mtrans(typeRNum *out, typeUSERPARAM *userparam)
        {
        }

        """
        g.auxiliaries += self.added_declarations


class SourceArtifact:
    def __init__(self, name):
        self.name = name



class ConstraintInspector:
    def __init__(self, method, stage):
        self.opti = Opti()

        self.X = self.opti.variable(*stage.x.shape)
        self.U = self.opti.variable(*stage.u.shape)
        self.V = self.opti.variable(*stage.v.shape)
        self.P = self.opti.parameter(*stage.p.shape)
        self.t = self.opti.parameter()
        self.T = self.opti.variable()

        self.raw = [stage.x,stage.u,stage.p,stage.t, method.v]
        self.optivar = [self.X, self.U, self.P, self.t, self.V]

        if method.free_time:
            self.raw += [stage.T]
            self.optivar += [self.T]
    
    def finalize(self):
        self.opti_advanced = self.opti.advanced

    def canon(self,expr):
        c = substitute([expr],self.raw,self.optivar)[0]
        mc = self.opti_advanced.canon_expr(c) # canon_expr should have a static counterpart
        return substitute([mc.lb,mc.canon,mc.ub],self.optivar,self.raw), mc

class GrampcMethod(ExternalMethod):
    def __init__(self,
        verbose=False,
        debug=False,
        grampc_options=None,
        **kwargs):
        """
        dt is post-processing: interplin
        

        GRAMPC is very much realtime iteration
        By default, ConvergenceCheck is even off, and you perform MaxMultIter outer iterations with MaxGradIter inner iterations.
        Default MaxMultIter = 1.

        
        """
        supported = {"free_T"}
        ExternalMethod.__init__(self, supported=supported, **kwargs)
        self.grampc_options = {} if grampc_options is None else grampc_options
        our_defaults = {"MaxMultIter": 3000, "ConvergenceCheck": "on", "MaxGradIter": 100, "ConstraintsAbsTol": 1e-8, "ConvergenceGradientRelTol": 1e-8,"LineSearchType": "adaptive"}
        for k,v in our_defaults.items():
            if k not in self.grampc_options:
                self.grampc_options[k] = v
        self.codegen_name = 'casadi_codegen'
        self.grampc_driver = 'grampc_driver'
        self.user = "((cs_struct*) userparam)"
        self.user_grampc = "((cs_struct*) grampc->userparam)"
        self.Nhor = self.N+1
        self.verbose = verbose
        self.debug = debug
        self.artifacts = []

    def fill_placeholders_integral(self, phase, stage, expr, *args):
        if phase==1:
            return expr

    def fill_placeholders_sum_control(self, phase, stage, expr, **kwargs):
        raise Exception("ocp.sum not supported. Use ocp.integral instead.")

    def fill_placeholders_sum_control_plus(self, phase, stage, expr, **kwargs):
        raise Exception("ocp.sum not supported. Use ocp.integral instead.")

    def _register(self,fun_name,argtypes,restype):
        self.prefix=""
        fun = getattr(self.lib,self.prefix+fun_name)
        setattr(self,"_"+fun_name,fun)
        fun.argtypes = argtypes
        fun.restype = restype

    def gen_interface(self, f):
        f = f.expand()
        self.codegen.add(f)
        self.preamble.append(f"{f.name()}_incref();")
        self.preamble.append(f"{f.name()}_work(&sz_arg_local, &sz_res_local, &sz_iw_local, &sz_w_local);")
        self.preamble.append("if (sz_arg_local>sz_arg) sz_arg=sz_arg_local;")
        self.preamble.append("if (sz_res_local>sz_res) sz_res=sz_res_local;")
        self.preamble.append("if (sz_iw_local>sz_iw) sz_iw=sz_iw_local;")
        self.preamble.append("if (sz_w_local>sz_iw) sz_w=sz_w_local;")
        self.postamble.append(f"{f.name()}_decref();")

        scalar = lambda name: name in ["t","T"]

        args = [
                f"ctypeRNum {'' if scalar(f.name_in(i)) else '*'}{f.name_in(i)}"
                    for i in range(f.n_in())
                    if "p_fixed" not in f.name_in(i)]
        self.wrapper.add_declarations(f"void {f.name()[3:]}(typeRNum *out, {', '.join(args)}, typeUSERPARAM *userparam) {{\n")
        self.wrapper.add_declarations("  int mem;\n")
        adj_i = None
        for i in range(f.n_in()):
            e = f.name_in(i)
            if "adj" in e:
                adj_i = i
            if scalar(e):
                self.wrapper.add_declarations(f"  {self.user}->arg[{i}] = &{e};\n")
            elif e=="p_fixed":
                self.wrapper.add_declarations(f"  {self.user}->arg[{i}] = {self.user}->p;\n")
            else:
                self.wrapper.add_declarations(f"  {self.user}->arg[{i}] = {e};\n")
        self.wrapper.add_declarations(f"  {self.user}->res[0] = out;\n")
        self.wrapper.add_declarations(f"  mem = {f.name()}_checkout();\n")
        self.wrapper.add_declarations(f"  {f.name()}({self.user}->arg, {self.user}->res, {self.user}->iw, {self.user}->w, mem);\n")
        self.wrapper.add_declarations(f"  {f.name()}_release(mem);\n")
        self.wrapper.add_declarations("}\n")

    def transcribe_phase1(self, stage, **kwargs):

        self.preamble = ["casadi_int sz_arg=0, sz_res=0, sz_iw=0, sz_w=0;",
                         "casadi_int sz_arg_local, sz_res_local, sz_iw_local, sz_w_local;",
                        ]
        self.postamble = []

        self.stage = stage


        f = stage._ode()
        options = {}
        options["with_header"] = True
        self.codegen = CodeGenerator(f"{self.codegen_name}.c", options)

        self.artifacts.append(SourceArtifact(f"{self.codegen_name}.c"))
        self.artifacts.append(SourceArtifact(f"{self.codegen_name}.h"))

        self.wrapper_codegen = MyCodeGenerator(self.grampc_driver)
        self.wrapper_codegen.add_include(f"{self.codegen_name}.h")
        self.wrapper_codegen.add_include("time.h")
        
        self.artifacts.append(SourceArtifact(f"{self.grampc_driver}.c"))
        self.wrapper = Wrapper(userparam=self.user_grampc)

        assert len(stage.variables['control'])==0, "variables defined on control grid not supported. Use controls instead."

        self.v = vvcat(stage.variables[''])
        self.X_gist = [MX.sym("Xg", stage.nx) for k in range(self.N+1)]
        self.U_gist = [MX.sym("Ug", stage.nu) for k in range(self.N)]
        self.V_gist = MX.sym("Vg", *self.v.shape)
        self.T_gist = MX.sym("Tg")

        assert f.numel_out("alg")==0
        assert f.numel_out("quad")==0
        ffct = Function("cs_ffct", [stage.t, stage.x, stage.u, self.v, stage.p], [ densify(f(x=stage.x, u=stage.u, p=stage.p, t=stage.t)["ode"])],['t','x','u','p','p_fixed'],['out'])
        self.gen_interface(ffct)
        self.gen_interface(ffct.factory("cs_dfdx_vec",["t","x","adj:out","u","p","p_fixed"],["densify:adj:x"]))
        self.gen_interface(ffct.factory("cs_dfdu_vec",["t","x","adj:out","u","p","p_fixed"],["densify:adj:u"]))
        self.gen_interface(ffct.factory("cs_dfdp_vec",["t","x","adj:out","u","p","p_fixed"],["densify:adj:p"]))

        self.constraint_inspector = ConstraintInspector(self, stage)

        #self.time_grid = self.grid(stage._t0, stage._T, self.N)
        self.normalized_time_grid = self.grid(0.0, 1.0, self.N)
        self.time_grid = self.normalized_time_grid
        if not isinstance(stage._T, FreeTime): self.time_grid*= stage._T
        if not isinstance(stage._t0, FreeTime): self.time_grid+= stage._t0
        self.control_grid = MX(stage.t0 + self.normalized_time_grid*stage.T).T

        inits = []
        inits.append((stage.T, stage._T.T_init if isinstance(stage._T, FreeTime) else stage._T))
        inits.append((stage.t0, stage._t0.T_init if isinstance(stage._t0, FreeTime) else stage._t0))

        self.control_grid_init = evalf(substitute([self.control_grid], [a for a,b in inits],[b for a,b in inits])[0])

        #self.control_grid = self.normalized_time_grid

        self.lagrange = MX(0)
        self.mayer = MX(0)
        var_mayer = []
        obj = MX(stage._objective)
        terms = get_terms(obj)
        for term in terms:
            n = [e.name() for e in symvar(term)]
            sumi = np.sum([e=="r_integral" for e in n])
            summ = np.sum([e=="r_at_tf" for e in n])
            if sumi>1:
                raise Exception("Objective cannot be parsed: operation combining two integrals not supported")
            if sumi==1:
                n_hits = 0
                for e,parents in visit(term):
                    if e.is_symbolic() and e.name()=="r_integral":
                        n_hits+=1
                        for pi in range(len(parents)):
                            p = parents[pi]
                            pchild = parents[pi-1] if pi>0 else e
                            correct = False
                            # Distributive operations
                            if p.op()==cs.OP_MUL:
                                correct = True
                            if p.op()==cs.OP_DIV:
                                # Only allow integral in LHS
                                correct = hash(p.dep(0))==hash(pchild)
                            assert correct, "Objective cannot be parsed: integrals can only be multiplied or divided."
                    if n_hits>1:
                        raise Exception("Objective cannot be parsed")
                if summ!=0:
                    raise Exception("Objective cannot be parsed: operation combining integral and at_tf part not supported")
                self.lagrange += term
                continue
            self.mayer += term
        self.P0 = DM.zeros(stage.np)

    def transcribe_phase2(self, stage, **kwargs):

        self.constraint_inspector.finalize()
        
        placeholders = kwargs["placeholders"]



        # Total Lagrange integrand
        lagrange = placeholders(self.lagrange,preference=['expose'])
        # Total Mayer term
        mayer = placeholders(self.mayer,preference=['expose'])

        xdes = MX.sym("xdes", stage.x.sparsity())
        udes = MX.sym("udes", stage.u.sparsity())
        lfct = Function("cs_lfct", [stage.t, stage.x, stage.u, self.v, stage.p, xdes, udes], [densify(lagrange)], ["t", "x", "u", "p", "p_fixed", "xdes", "udes"], ["out"])
        self.gen_interface(lfct)
        self.gen_interface(lfct.factory("cs_dldx",["t","x","u","p","p_fixed", "xdes", "udes"],["grad:out:x"]))
        self.gen_interface(lfct.factory("cs_dldu",["t","x","u","p","p_fixed", "xdes", "udes"],["grad:out:u"]))
        self.gen_interface(lfct.factory("cs_dldp",["t","x","u","p","p_fixed", "xdes", "udes"],["grad:out:p"]))

        Vfct = Function("cs_Vfct", [stage.T, stage.x, self.v, stage.p, xdes], [densify(mayer)], ["T", "x", "p", "p_fixed", "xdes"], ["out"])

        self.gen_interface(Vfct)
        self.gen_interface(Vfct.factory("cs_dVdx",["T","x","p","p_fixed","xdes"],["grad:out:x"]))
        self.gen_interface(Vfct.factory("cs_dVdp",["T","x","p","p_fixed","xdes"],["grad:out:p"]))
        self.gen_interface(Vfct.factory("cs_dVdT",["T","x","p","p_fixed","xdes"],["grad:out:T"]))

        eq = [] #
        ineq = [] # <=0
        eq_term = []
        ineq_term = []

        # helpers to put limits on u
        ub_expr = []
        ub_l = []
        ub_u = []



        # Process path constraints
        for c, meta, args in stage._constraints["control"]+stage._constraints["integrator"]:
            (lb,canon,ub),mc = self.constraint_inspector.canon(placeholders(c,preference=['expose']))
            # lb <= canon <= ub
            # Check for infinities
            try:
                lb_inf = np.all(np.array(evalf(lb)==-inf))
            except:
                lb_inf = False
            try:
                ub_inf = np.all(np.array(evalf(ub)==inf))
            except:
                ub_inf = False

            if mc.type == casadi.OPTI_EQUALITY:
                eq.append(canon-ub)
            else:
                assert mc.type in [casadi.OPTI_INEQUALITY, casadi.OPTI_GENERIC_INEQUALITY, casadi.OPTI_DOUBLE_INEQUALITY]

                # Catch simple bounds on u
                if is_linear(canon, stage.u) and not depends_on(canon, vertcat(stage.x, self.v)):
                    J,c = linear_coeff(canon, stage.u)
                    try:
                        check_Js(J)
                        ub_expr.append(J)
                        if ub_inf:
                            ub_u.append(reshape_number(J @ stage.u,-INF))
                        else:
                            ub_u.append(ub-c)
                        if lb_inf:
                            ub_l.append(reshape_number(J @ stage.u,INF))
                        else:
                            ub_l.append(lb-c)
                        continue
                    except:
                        pass

                if not ub_inf:
                    ineq.append(canon-ub)
                if not lb_inf:
                    ineq.append(lb-canon)

        eq = vvcat(eq)
        ineq = vvcat(ineq)

        ub_expr = evalf(vcat(ub_expr))
        ub_l = evalf(vcat(ub_l))
        ub_u = evalf(vcat(ub_u))
        # Add missing rows
        rows = set(sum1(ub_expr).T.row())
        missing_rows = [i for i in range(stage.nu) if i not in rows]
        M = DM(len(missing_rows), stage.nu)
        for i,e in enumerate(missing_rows):
            M[i,e] = 1

        ub_expr = vertcat(ub_expr,M)
        ub_l = vertcat(ub_l,-INF*DM.ones(len(missing_rows)))
        ub_u = vertcat(ub_u,INF*DM.ones(len(missing_rows)))
       
        ub_l = solve(ub_expr,ub_l)
        ub_u = solve(ub_expr,ub_u)

        # No export_num here, let's do things parametrically
        self.m = m = OrderedDict()

        m["umin"] = export_vec(ub_l)
        m["umax"] = export_vec(ub_u)

        gfct = Function("cs_gfct", [stage.t, stage.x, stage.u, self.v, stage.p], [densify(eq)], ["t", "x", "u", "p", "p_fixed"], ["out"])
        self.gen_interface(gfct)
        self.gen_interface(gfct.factory("cs_dgdx_vec",["t", "x", "u", "p", "adj:out", "p_fixed"],["densify:adj:x"]))
        self.gen_interface(gfct.factory("cs_dgdu_vec",["t", "x", "u", "p", "adj:out", "p_fixed"],["densify:adj:u"]))
        self.gen_interface(gfct.factory("cs_dgdp_vec",["t", "x", "u", "p", "adj:out", "p_fixed"],["densify:adj:p"]))

        hfct = Function("cs_hfct", [stage.t, stage.x, stage.u, self.v, stage.p], [densify(ineq)], ["t", "x", "u", "p", "p_fixed"], ["out"])
        self.gen_interface(hfct)
        self.gen_interface(hfct.factory("cs_dhdx_vec",["t", "x", "u", "p", "adj:out", "p_fixed"],["densify:adj:x"]))
        self.gen_interface(hfct.factory("cs_dhdu_vec",["t", "x", "u", "p", "adj:out", "p_fixed"],["densify:adj:u"]))
        self.gen_interface(hfct.factory("cs_dhdp_vec",["t", "x", "u", "p", "adj:out", "p_fixed"],["densify:adj:p"]))


        x0_eq = []
        x0_b = []

        Tmin = 0
        Tmax = INF

        # Process point constraints
        # Probably should de-duplicate stuff wrt path constraints code
        for c, meta, _ in stage._constraints["point"]:
            # Make sure you resolve u to r_at_t0/r_at_tf
            c = placeholders(c,max_phase=1)
            has_t0 = 'r_at_t0' in [a.name() for a in symvar(c)]
            has_tf = 'r_at_tf' in [a.name() for a in symvar(c)]

            cb = c
            (lb,canon,ub),mc = self.constraint_inspector.canon(placeholders(c,preference=['expose']))

            if has_t0:
                # t0
                check = is_linear(canon, stage.x)
                check = check and not depends_on(canon, vertcat(stage.u, self.v))
                assert check and mc.type == casadi.OPTI_EQUALITY, "at t=t0, only equality constraints on x are allowed. Got '%s'" % str(c)

                J,c = linear_coeff(canon, stage.x)
                try:
                    J = evalf(J)
                    x0_eq.append(J)
                    x0_b.append(lb-c)
                    continue
                except:
                    pass

            # Check for infinities
            try:
                lb_inf = np.all(np.array(evalf(lb)==-inf))
            except:
                lb_inf = False
            try:
                ub_inf = np.all(np.array(evalf(ub)==inf))
            except:
                ub_inf = False

            if mc.type == casadi.OPTI_EQUALITY:
                eq_term.append(canon-ub)
            else:
                assert mc.type in [casadi.OPTI_INEQUALITY, casadi.OPTI_GENERIC_INEQUALITY, casadi.OPTI_DOUBLE_INEQUALITY]
                # Catch simple bounds on T
                if self.free_time:
                    if is_linear(canon, stage.T) and not depends_on(canon, vertcat(stage.x, stage.u, self.v)):
                        J,c = linear_coeff(canon, stage.T)
                        if not ub_inf:
                            Tmax = fmin(Tmax, (ub-c)/J)
                        if not lb_inf:
                            Tmin = fmax(Tmin, (lb-c)/J)
                        continue

                if not ub_inf:
                    ineq_term.append(canon-ub)
                if not lb_inf:
                    ineq_term.append(lb-canon)

        Tmin = fmax(Tmin, 1/INF)
        m["Tmin"] = export(Tmin)
        m["Tmax"] = export(Tmax)

        x0_eq = vcat(x0_eq)
        x0_b = vcat(x0_b)
        x0_expr = casadi.inv(evalf(x0_eq)) @ x0_b # casadi.solve(x0_eq, x0_b)
        m["x_current"] = export_vec(x0_expr)

        eq_term = vvcat(eq_term)
        ineq_term = vvcat(ineq_term)

        gTfct = Function("cs_gTfct", [stage.T, stage.x, self.v, stage.p], [densify(eq_term)], ["T", "x", "p", "p_fixed"], ["out"])
        gTfct.disp(True)
        self.gen_interface(gTfct)
        self.gen_interface(gTfct.factory("cs_dgTdx_vec",["T", "x", "p", "adj:out", "p_fixed"],["densify:adj:x"]))
        self.gen_interface(gTfct.factory("cs_dgTdp_vec",["T", "x", "p", "adj:out", "p_fixed"],["densify:adj:p"]))
        self.gen_interface(gTfct.factory("cs_dgTdT_vec",["T", "x", "p", "adj:out", "p_fixed"],["densify:adj:T"]))

        hTfct = Function("cs_hTfct", [stage.T, stage.x, self.v, stage.p], [densify(ineq_term)], ["T", "x", "p", "p_fixed"], ["out"])
        hTfct.disp(True)
        self.gen_interface(hTfct)
        self.gen_interface(hTfct.factory("cs_dhTdx_vec",["T", "x", "p", "adj:out", "p_fixed"],["densify:adj:x"]))
        self.gen_interface(hTfct.factory("cs_dhTdp_vec",["T", "x", "p", "adj:out", "p_fixed"],["densify:adj:p"]))
        self.gen_interface(hTfct.factory("cs_dhTdT_vec",["T", "x", "p", "adj:out", "p_fixed"],["densify:adj:T"]))

        args = [v[0] for v in m.values()]
        self.pmap = Function('pmap',[stage.p],args,['p'],list(m.keys()))
        self.codegen.add(self.pmap)

        self.U0 = U0 = DM.zeros(stage.nu)
        self.V0 = V0 = DM.zeros(self.v.numel())


        for var, expr in stage._initial.items():
            assert not depends_on(expr, stage.t)
            assert not depends_on(var, stage.x)
            if depends_on(var,stage.u):
                assert not depends_on(var, self.v)
                J, r = linear_coeffs(var,stage.u)
                J = evalf(J)
                r = evalf(r)
                assert r.is_zero()
                check_Js(J)
                expr = reshape_number(var, expr)
                if J.sparsity().get_col():
                    U0[J.sparsity().get_col()] = expr[J.row()]
            else:
                assert not depends_on(var, stage.u)
                J, r = linear_coeffs(var,self.v)
                J = evalf(J)
                r = evalf(r)
                assert r.is_zero()
                check_Js(J)
                expr = reshape_number(var, expr)
                if J.sparsity().get_col():
                    V0[J.sparsity().get_col()] = expr[J.row()]
        self.wrapper.add_declarations(
        f"""
            /** OCP dimensions: states (Nx), controls (Nu), parameters (Np), equalities (Ng), 
            inequalities (Nh), terminal equalities (NgT), terminal inequalities (NhT) **/
            void ocp_dim(typeInt *Nx, typeInt *Nu, typeInt *Np, typeInt *Ng, typeInt *Nh, typeInt *NgT, typeInt *NhT, typeUSERPARAM *userparam)
            {{
                *Nx = {stage.nx};
                *Nu = {stage.nu};
                *Np = {self.v.numel()};
                *Ng = {eq.numel()};
                *Nh = {ineq.numel()};
                *NgT = {eq_term.numel()};
                *NhT = {ineq_term.numel()};
            }}
        """)


        self.wrapper.add_declarations("void preamble(typeUSERPARAM* userparam) {\n")
        for l in self.preamble:
            self.wrapper.add_declarations("  " + l + "\n")
        self.wrapper.add_declarations(f"""
        {self.user}->arg = malloc(sizeof(const casadi_real*)*sz_arg);
        {self.user}->res = malloc(sizeof(casadi_real*)*sz_res);
        {self.user}->iw = sz_iw>0 ? malloc(sizeof(casadi_int)*sz_iw) : 0;
        {self.user}->w = sz_w>0 ? malloc(sizeof(casadi_real)*sz_w) : 0;
        {self.user}->x_opt = malloc(sizeof(casadi_real)*{max(stage.nx*self.Nhor,1)});
        {self.user}->u_opt = malloc(sizeof(casadi_real)*{max(stage.nu*self.Nhor,1)});
        {self.user}->v_opt = malloc(sizeof(casadi_real)*{max(self.v.numel(),1)});
        {self.user}->x_current = malloc(sizeof(casadi_real)*{max(stage.nx, 1)});
        {self.user}->p = malloc(sizeof(casadi_real)*{max(stage.np, 1)});
        {self.user}->umin = malloc(sizeof(casadi_real)*{max(stage.nu, 1)});
        {self.user}->umax = malloc(sizeof(casadi_real)*{max(stage.nu, 1)});
        {self.user}->u0 = malloc(sizeof(casadi_real)*{max(stage.nu, 1)});
        {self.user}->v0 = malloc(sizeof(casadi_real)*{max(self.v.numel(), 1)});
        """)
        self.wrapper.add_declarations("}\n")

        self.wrapper.add_declarations("void postamble(typeUSERPARAM* userparam) {\n")
        for l in self.postamble:
            self.wrapper.add_declarations("  " + l + "\n")
        self.wrapper.add_declarations(f"""
        free({self.user}->arg);
        free({self.user}->res);
        free({self.user}->iw);
        free({self.user}->w);
        free({self.user}->x_opt);
        free({self.user}->u_opt);
        free({self.user}->v_opt);
        free({self.user}->x_current);
        free({self.user}->p);
        free({self.user}->umin);
        free({self.user}->umax);
        free({self.user}->u0);
        free({self.user}->v0);
        }}
        """)

        nc = vertcat(eq,ineq,eq_term,ineq_term).numel()
        vector_options = {"ConstraintsAbsTol": nc}
        for k, L in sorted(vector_options.items()):
            if k in self.grampc_options:
                if isinstance(self.grampc_options[k],float):
                    self.grampc_options[k] = [self.grampc_options[k]]*L
                self.wrapper.add_init_mem(f"""double {k}[{L}] = {{{strlist(self.grampc_options[k])}}};\n""")


        for k,v in stage._param_vals.items():
            self.set_value(stage, self, k, v)

        res = self.pmap(p=self.P0)

        p = self.P0.nonzeros()
        x_current = res["x_current"].nonzeros()
        umax = res["umax"].nonzeros()
        umin = res["umin"].nonzeros()
        u0 = U0.nonzeros()
        v0 = V0.nonzeros()
        Tmin = float(res["Tmin"])
        Tmax = float(res["Tmax"])
        def brace(a): return "{"+a+"}"
        self.wrapper.add_init_mem(f"""
            int i;
            typeUSERPARAM* userparam = malloc(sizeof(cs_struct));
            double x0[{max(stage.nx,1)}] = {{{strlist(x_current)}}};
            double umax[{max(stage.nu,1)}] = {{{strlist(umax)}}};
            double umin[{max(stage.nu,1)}] = {{{strlist(umin)}}};
            double u0[{max(stage.nu,1)}] = {{{strlist(u0)}}};
            double p[{max(stage.np,1)}]{ " = "+ brace(strlist(p)) if p else ""};
            double v0[{max(self.v.numel(),1)}]{ " = "+ brace(strlist(v0)) if v0 else ""};
            preamble(userparam);

            for (i=0;i<{stage.np};++i) {self.user}->p[i] = p[i];

            /********* grampc init *********/
            grampc_init(&grampc, userparam);

            grampc_setparam_real_vector(grampc, "x0", x0);
            grampc_setparam_real_vector(grampc, "umax", umax);
            grampc_setparam_real_vector(grampc, "umin", umin);
            grampc_setparam_real_vector(grampc, "u0", u0);
            grampc_setparam_real_vector(grampc, "p0", v0);

            //grampc_setparam_real_vector(grampc, "xdes", 0);
            //grampc_setparam_real_vector(grampc, "udes", 0);

            grampc_setparam_real(grampc, "Thor", {self.control_grid_init[-1]-self.control_grid_init[0]});

            grampc_setparam_real(grampc, "dt", {self.control_grid_init[1]-self.control_grid_init[0]});
            grampc_setparam_real(grampc, "t0", {self.control_grid_init[0]});

            grampc_setopt_int(grampc, "Nhor", {self.Nhor});

            /********* Option definition *********/

            grampc_setopt_string(grampc, "OptimTime", "{"on" if self.free_time else "off"}");
            grampc_setopt_string(grampc, "OptimParam", "{"on" if self.v.numel()>0 else "off"}");
            """)

        if self.free_time:
            self.wrapper.add_init_mem(f"""
                grampc_setparam_real(grampc, "Tmin", {Tmin});
                grampc_setparam_real(grampc, "Tmax", {Tmax});
            """)

        int_options = ["MaxGradIter","MaxMultIter","Nhor","IntegratorMaxSteps"]
        for k,v in sorted(self.grampc_options.items()):
            if k in vector_options.keys():
                self.wrapper.add_init_mem(f"grampc_setopt_real_vector(grampc, \"{k}\", {k});\n")
            if k in int_options:
                self.wrapper.add_init_mem(f"grampc_setopt_int(grampc, \"{k}\", {v});\n")
            elif isinstance(v, float) or isinstance(v, int):
                self.wrapper.add_init_mem(f"grampc_setopt_real(grampc, \"{k}\", {v});\n")
            elif isinstance(v, str):
                self.wrapper.add_init_mem(f"grampc_setopt_string(grampc, \"{k}\", \"{v}\");\n")
        if self.verbose:
            self.wrapper.add_init_mem(f"""
                grampc_printopt(grampc);
                grampc_printparam(grampc);
            """)

        self.wrapper.add_init_mem(f"""
            /********* estimate and set PenaltyMin *********/
            grampc_estim_penmin(grampc, 1);
        """)
        if self.verbose:
            self.wrapper.add_init_mem(f"""
                grampc_printopt(grampc);
                grampc_printparam(grampc);
            """)

        grampc = self.wrapper.codegen_mem(self.wrapper_codegen)
        stats = self.user_grampc+"->stats"
        self.wrapper.add_declarations(f"""
            const grampc_solver_stats* grampc_driver_get_stats_internal(void) {{
                typeGRAMPC *grampc = {grampc};
                return &{stats};
            }}

            const compat_solver_stats* grampc_driver_get_stats(void) {{
                typeGRAMPC *grampc = {grampc};
                static compat_solver_stats ret;
                ret.n_sqp_iter = {stats}.n_outer_iter;
                ret.n_qp_iter = {stats}.n_inner_iter;
                ret.sqp_stop_crit = {stats}.conv_grad + 2*{stats}.conv_con;
                ret.runtime = {stats}.runtime;
                return &ret;
            }}
        
        """)


        self.wrapper.set_sp_in([Sparsity.dense(stage.np)])
        self.wrapper.set_name_in(["p"])


        self.wrapper.set_sp_out([Sparsity.dense(stage.nx, self.N+1), Sparsity.dense(stage.nu, self.N), Sparsity.dense(self.v.numel()), Sparsity.dense(1,1)])
        self.wrapper.set_name_out(["x_opt","u_opt","v_opt","T_opt"])

 
        self.wrapper.add_body(f"""
            int mm;
            clock_t start_t, end_t;
            start_t=clock();
            const casadi_real ** pmap_arg = arg+{self.wrapper.n_in()};
            casadi_real ** pmap_res = res+{self.wrapper.n_out()};
            pmap_arg[0] = arg[0];
            

            for (int i=0;i<{stage.np};++i) {self.user_grampc}->p[i] = pmap_arg[0][i];
            """)
        for i in range(self.pmap.n_out()):
            n = self.pmap.name_out(i)
            self.wrapper.add_body(f"""pmap_res[{i}] = {"&" if n.startswith("T") else ""}{self.user_grampc}->{n};\n""")

        def lookup_out(name):
            return "res[" + str(["x_opt","u_opt","v_opt","T_opt"].index(name))+"]"

        self.wrapper.add_body(f"""
            mm = pmap_checkout();
            pmap(pmap_arg, pmap_res, iw, w, 0);
            pmap_release(mm);

            grampc_setparam_real_vector(grampc, "x0", {self.user_grampc}->x_current);
            grampc_setparam_real_vector(grampc, "umin", {self.user_grampc}->umin);
            grampc_setparam_real_vector(grampc, "umax", {self.user_grampc}->umax);
            grampc_setparam_real(grampc, "Tmin", {self.user_grampc}->Tmin);
            grampc_setparam_real(grampc, "Tmax", {self.user_grampc}->Tmax);

            grampc_run(grampc);

            end_t = clock();

            {self.user_grampc}->stats.runtime = (casadi_real)(end_t - start_t) / CLOCKS_PER_SEC;
            
            """)
        
        if self.verbose:
            self.wrapper.add_body(f"grampc_printstatus(grampc->sol->status, STATUS_LEVEL_DEBUG);")
        self.wrapper.add_body(f"""
            for(int k=0;k<grampc->opt->Nhor;++k) {{
                for (int j=0;j<{stage.nx};++j) {{
                    int i = k*{stage.nx}+j;
                    {self.user_grampc}->x_opt[i] = grampc->rws->x[i]*grampc->opt->xScale[j]+grampc->opt->xOffset[j];
                }}
                for (int j=0;j<{stage.nu};++j) {{
                    int i = k*{stage.nu}+j;
                    {self.user_grampc}->u_opt[i] = grampc->rws->u[i]*grampc->opt->uScale[j]+grampc->opt->uOffset[j];
                }}
            }}
            for (int i=0;i<{self.v.numel()};++i) {{
                {self.user_grampc}->v_opt[i] = grampc->rws->p[i]*grampc->opt->pScale[i]+grampc->opt->pOffset[i];
            }}
            {self.user_grampc}->T_opt = grampc->rws->T;

            if ({lookup_out("x_opt")}) for (int i=0;i<{stage.nx*self.Nhor};++i) {lookup_out("x_opt")}[i] = {self.user_grampc}->x_opt[i];
            if ({lookup_out("u_opt")}) for (int i=0;i<{stage.nu*self.N};++i) {lookup_out("u_opt")}[i] = {self.user_grampc}->u_opt[i];
            if ({lookup_out("v_opt")}) for (int i=0;i<{self.v.numel()};++i) {lookup_out("v_opt")}[i] = {self.user_grampc}->v_opt[i];
            if ({lookup_out("T_opt")}) {lookup_out("T_opt")}[0] = {self.user_grampc}->T_opt;

            for ({self.user_grampc}->stats.n_outer_iter=0,{self.user_grampc}->stats.n_inner_iter=0;{self.user_grampc}->stats.n_outer_iter<grampc->opt->MaxMultIter;++{self.user_grampc}->stats.n_outer_iter) {{
                int n = grampc->sol->iter[{self.user_grampc}->stats.n_outer_iter];
                {self.user_grampc}->stats.n_inner_iter += n;
                if (n==0) break;
            }}

            {self.user_grampc}->stats.conv_grad = convergence_test_gradient(grampc->opt->ConvergenceGradientRelTol, grampc);
            {self.user_grampc}->stats.conv_con = convergence_test_constraints(grampc->opt->ConstraintsAbsTol, grampc);


        """)
        
        build_dir_abs = "foobar"
        self.build_dir_abs = build_dir_abs
        prepare_build_dir(build_dir_abs)

        self.codegen.generate(build_dir_abs+os.sep)

        self.wrapper_codegen.add_dependency(self.wrapper)
        self.wrapper_codegen.generate(build_dir_abs+os.sep)

        
        for e in glob.glob(f"{external_dir}/include/*.h"):
            self.artifacts.append(SourceArtifact(e))
            shutil.copy(e,build_dir_abs)

        for e in glob.glob(f"{external_dir}/src/*.c"):
            self.artifacts.append(SourceArtifact(e))
            shutil.copy(e,build_dir_abs)

        # for e in glob.glob(f"{external_dir}/libs/*"):
        #    shutil.copy(e,build_dir_abs)

        cmake_file_name = os.path.join(build_dir_abs,"CMakeLists.txt")
        with open(cmake_file_name,"w") as out:
            out.write(f"""
            project(grampc_export)

            set(CMAKE_WINDOWS_EXPORT_ALL_SYMBOLS ON)
            SET(CMAKE_INSTALL_RPATH_USE_LINK_PATH TRUE)
            SET(CMAKE_INSTALL_RPATH "$ORIGIN")

            cmake_minimum_required(VERSION 3.0)

            {'set(CMAKE_BUILD_TYPE Debug)' if self.debug else ''}

            #add_library(grampc INTERFACE)

            #find_library(GRAMPC_LIB NAMES grampc REQUIRED HINTS .)
            #find_path(GRAMPC_INCLUDE_DIR grampc.h HINTS .)

            #target_link_libraries(grampc INTERFACE ${{GRAMPC_LIB}} m)
            #target_include_directories(grampc INTERFACE ${{GRAMPC_INCLUDE_DIR}})

            add_library({self.grampc_driver} SHARED {self.grampc_driver}.c {self.codegen_name}.c
            
            euler1.c     grampc_alloc.c      grampc_init.c  grampc_run.c     grampc_setparam.c  heun2.c  ruku45.c   trapezodial.c
eulermod2.c  grampc_fixedsize.c  grampc_mess.c  grampc_setopt.c  grampc_util.c      rodas.c  simpson.c
euler1.h     f2cmod.h        grampc_fixedsize.h  grampc_init.h   grampc_mess.h  grampc_setopt.h    grampc_util.h  probfct.h           rodas.h   simpson.h
eulermod2.h  grampc_alloc.h  grampc.h            grampc_macro.h  grampc_run.h   grampc_setparam.h  heun2.h        rodas_decsol_f2c.h  ruku45.h  trapezodial.h
            )
            
            #target_link_libraries({self.grampc_driver} grampc)

            if (UNIX)
            target_link_libraries({self.grampc_driver} m)
            endif()

            install(TARGETS {self.grampc_driver} RUNTIME DESTINATION . LIBRARY DESTINATION .)

            """)
        script_file_name = os.path.join(build_dir_abs,"build.bat")
        with open(script_file_name,"w") as out:
            out.write(f"""
            echo "Should be ran in 'x64 Native Tools Command Prompt for VS'"
            mkdir build
            cd build && cmake -G "{windows_generator}" -A x64 -DCMAKE_INSTALL_PREFIX=.. .. && cd ..
            cmake --build build --config Release --target install
            """)
        script_file_name = os.path.join(build_dir_abs,"build.sh")
        with open(script_file_name,"w") as out:
            out.write(f"""
            mkdir -p build
            cd build && cmake -DCMAKE_INSTALL_PREFIX=.. .. && cd ..
            cmake --build build --target install
            """)
        #subprocess.run(["bash","build.sh"],cwd=build_dir_abs)
        run_build(cwd=build_dir_abs)

        if os.name == "nt":
            libname = self.grampc_driver+".dll"
        else:
            libname = "lib"+self.grampc_driver+".so"

        libname_full = os.path.join(build_dir_abs,libname)
        libname_extra = hashlib.md5(open(libname_full,'rb').read()).hexdigest()[:8]+libname
        libname_extra_full = os.path.join(build_dir_abs,libname_extra)
        shutil.copy(libname_full,libname_extra_full)
 
        self.solver = external(self.grampc_driver, libname_extra_full)

        if self.debug:
            cg = CodeGenerator("debug",{"main":True})
            solver_wrapper = self.solver.wrap()
            cg.add(solver_wrapper)
            cg.generate(build_dir_abs+os.sep)
            with open(cmake_file_name,"a") as out:
                out.write(f"""
                add_executable(debug debug.c)
                target_link_libraries(debug {self.grampc_driver})

                install(TARGETS debug RUNTIME DESTINATION .)

                """)
            run_build(cwd=build_dir_abs)

        # PyDLL instead of CDLL to keep GIL:
        # virtual machine emits Python prints
        self.lib = PyDLL(libname_extra_full)
        self._register("grampc_driver_get_stats_internal",[], POINTER(StatsStruct))

    def to_function(self, stage, name, args, results, *margs):
        print("args=",args)

        res = self.solver(p=stage.p)
        print(stage.p)
        print([stage.value(a) for a in args])


        [_,states] = stage.sample(stage.x,grid='control')
        [_,controls] = stage.sample(stage.u,grid='control-')
        variables = stage.value(vvcat(stage.variables['']))

        helper_in = [states,controls,variables, stage.T]
        helper = Function("helper", helper_in, results)

        arg_in = helper(res["x_opt"],res["u_opt"],res["v_opt"],res["T_opt"])

        ret = Function(name, args, arg_in, *margs)
        assert not ret.has_free()
        return ret

    def initial_value(self, stage, expr):
        ret = self.pmap(p=self.P0)
        parameters = []
        for p in stage.parameters['']:
            parameters.append(stage.value(p))
        
        [_,states] = stage.sample(stage.x,grid='control')
        [_,controls] = stage.sample(stage.u,grid='control-')
        variables = stage.value(vvcat(stage.variables['']))

        helper_in = [vvcat(parameters),states,controls,variables, stage.T]
        helper = Function("helper", helper_in, [expr])
        return helper(self.P0, cs.repmat(ret["x_current"], 1, self.N+1), cs.repmat(self.U0, 1, self.N), self.V0, 0).toarray(simplify=True)

    def solve(self, stage,limited=False):
        self.solver.generate_in(self.build_dir_abs+os.sep+"debug_in.txt",[self.P0])
        ret = self.solver(p=self.P0)


        self.stats = stats = self._grampc_driver_get_stats_internal().contents
        self.last_solution = OcpSolution(SolWrapper(self, vec(ret["x_opt"]), vec(ret["u_opt"]), ret["v_opt"], ret["T_opt"], rT=stage.T), stage)

        conv = stats.conv_grad and stats.conv_con

        if not conv:
            if stats.n_outer_iter==self.grampc_options["MaxMultIter"]:
                if not limited:
                    raise Exception("MaxMultIter exhausted without meeting convergence criteria")
            else:
                raise Exception("Problem not converged")

        return self.last_solution

    def get_stats(self):
        stats = self.stats
        return dict((k,getattr(stats,k)) for k,_ in stats_fields)

    def non_converged_solution(self, stage):
        return self.last_solution

    def solve_limited(self, stage):
        return self.solve(stage,limited=True)

    def eval(self, stage, expr):
        placeholders = stage.placeholders_transcribed
        expr = placeholders(expr,max_phase=1)
        ks = [self.v,stage.T]
        vs = [self.V_gist, self.T_gist]
        ret = substitute([expr],ks,vs)[0]
        return ret
        
    @property
    def gist(self):
        return vertcat(ExternalMethod.gist.fget(self), self.V_gist, self.T_gist)

    def eval_at_control(self, stage, expr, k):
        placeholders = stage.placeholders_transcribed
        expr = placeholders(expr,max_phase=1)
        ks = [stage.x,stage.u,self.v,stage.T]
        vs = [self.X_gist[k], self.U_gist[min(k, self.N-1)], self.V_gist, self.T_gist]
        if not self.t_state:
            ks += [stage.t]
            vs += [self.control_grid[k]]
        ret = substitute([expr],ks,vs)[0]
        return ret

class SolWrapper:
    def __init__(self, method, x, u, v, T, rT=None):
        self.method = method
        self.x = x
        self.u = u
        self.T = T
        self.v = v
        self.rT = rT

    def value(self, expr, *args,**kwargs):
        placeholders = self.method.stage.placeholders_transcribed
        expr = substitute(expr,self.rT, self.T)
        ret = evalf(substitute([placeholders(expr)],[self.method.gist],[vertcat(self.x, self.u, self.v, self.T)])[0])
        return ret.toarray(simplify=True)

    def stats(self):
        return self.method.get_stats()



# -*- coding: future_fstrings -*-
#
# Copyright (c) The acados authors.
#
# This file is part of acados.
#
# The 2-Clause BSD License
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.;
#

import numpy as np
from .utils import check_if_nparray_and_flatten, check_if_2d_nparray

class AcadosOcpCost:
    r"""
    Class containing the numerical data of the cost:

    NOTE: By default, the Lagrange cost term provided in continuous time is internally integrated using the explicit Euler method, cost_discretization = 'EULER',
    which allows for a seamless OCP discretization with a nonuniform time grid.
    This means that all cost terms, except for the terminal one, are weighted with the corresponding time step.
    :math:`c_\text{total} = \Delta t_0 \cdot l_0(x_0, u_0, z_0, p_0) + ... + \Delta t_{N-1} \cdot l_{N-1}(x_{N-1}, u_{N-1}, z_{N-1}, p_{N-1}) + l_N(x_N, p_N)`.

    If a nonlinear least-squares or convex-over-nonlinear cost is used, the cost can also be integrated using the same integration scheme,
    which is used for the dynamics, cost_discretization = 'INTEGRATOR'.

    In case of LINEAR_LS:
    stage cost is
    :math:`l(x,u,z) = 0.5 \cdot || V_x \, x + V_u \, u + V_z \, z - y_\text{ref}||^2_W`,
    terminal cost is
    :math:`m(x) = 0.5 \cdot || V^e_x \, x - y_\text{ref}^e||^2_{W^e}`

    In case of NONLINEAR_LS:
    stage cost is
    :math:`l(x,u,z,t,p) = 0.5 \cdot || y(x,u,z,t,p) - y_\text{ref}||^2_W`,
    terminal cost is
    :math:`m(x,p) = 0.5 \cdot || y^e(x,p) - y_\text{ref}^e||^2_{W^e}`

    In case of CONVEX_OVER_NONLINEAR:
    stage cost is
    :math:`l(x,u,z,t,p) = \psi(y(x,u,z,t,p) - y_\text{ref}, t, p)`,
    terminal cost is
    :math:`m(x, p) = \psi^e (y^e(x,p) - y_\text{ref}^e, p)`
    """
    def __init__(self):
        # initial stage
        self.__cost_type_0 = None
        self.__W_0 = None
        self.__Vx_0 = None
        self.__Vu_0 = None
        self.__Vz_0 = None
        self.__yref_0 = None
        self.__Zl_0 = None
        self.__Zu_0 = None
        self.__zl_0 = None
        self.__zu_0 = None
        self.__cost_ext_fun_type_0 = 'casadi'
        self.__cost_source_ext_cost_0 = None # TODO add property, only required for generic
        self.__cost_function_ext_cost_0 = None # TODO add property, only required for generic

        # Lagrange term
        self.__cost_type   = 'LINEAR_LS'  # cost type
        self.__W           = np.zeros((0,0))
        self.__Vx          = np.zeros((0,0))
        self.__Vu          = np.zeros((0,0))
        self.__Vz          = np.zeros((0,0))
        self.__yref        = np.array([])
        self.__Zl          = np.array([])
        self.__Zu          = np.array([])
        self.__zl          = np.array([])
        self.__zu          = np.array([])

        # TODO: check how generic works in templates ?!
        self.__cost_ext_fun_type = 'casadi'
        self.__cost_source_ext_cost = None # TODO add property, only required for generic
        self.__cost_function_ext_cost = None # TODO add property, only required for generic

        # Mayer term
        self.__cost_type_e = 'LINEAR_LS'
        self.__W_e         = np.zeros((0,0))
        self.__Vx_e        = np.zeros((0,0))
        self.__yref_e      = np.array([])
        self.__Zl_e        = np.array([])
        self.__Zu_e        = np.array([])
        self.__zl_e        = np.array([])
        self.__zu_e        = np.array([])
        self.__cost_ext_fun_type_e = 'casadi'
        self.__cost_source_ext_cost_e = None # TODO add property, only required for generic
        self.__cost_function_ext_cost_e = None # TODO add property, only required for generic


    # initial stage
    @property
    def cost_type_0(self):
        """Cost type at initial shooting node (0)
        -- string in {EXTERNAL, LINEAR_LS, NONLINEAR_LS, CONVEX_OVER_NONLINEAR} or :code:`None`.
        Default: :code:`None`.

            .. note:: Cost at initial stage is the same as for intermediate shooting nodes if not set differently explicitly.

            .. note:: If :py:attr:`cost_type_0` is set to :code:`None` values in :py:attr:`W_0`, :py:attr:`Vx_0`, :py:attr:`Vu_0`, :py:attr:`Vz_0` and :py:attr:`yref_0` are ignored (set to :code:`None`).
        """
        return self.__cost_type_0

    @property
    def W_0(self):
        """:math:`W_0` - weight matrix at initial shooting node (0).
        Default: :code:`None`.
        """
        return self.__W_0

    @property
    def Vx_0(self):
        """:math:`V_x^0` - x matrix coefficient at initial shooting node (0).
        Default: :code:`None`.
        """
        return self.__Vx_0

    @property
    def Vu_0(self):
        """:math:`V_u^0` - u matrix coefficient at initial shooting node (0).
        Default: :code:`None`.
        """
        return self.__Vu_0

    @property
    def Vz_0(self):
        """:math:`V_z^0` - z matrix coefficient at initial shooting node (0).
        Default: :code:`None`.
        """
        return self.__Vz_0

    @property
    def yref_0(self):
        r""":math:`y_\text{ref}^0` - reference at initial shooting node (0).
        Default: :code:`None`.
        """
        return self.__yref_0

    @property
    def cost_ext_fun_type_0(self):
        """Type of external function for cost at initial shooting node (0)
        -- string in {casadi, generic} or :code:`None`
        Default: :code:'casadi'.

            .. note:: Cost at initial stage is the same as for intermediate shooting nodes if not set differently explicitly.
        """
        return self.__cost_ext_fun_type_0

    @yref_0.setter
    def yref_0(self, yref_0):
        yref_0 = check_if_nparray_and_flatten(yref_0, "yref_0")
        self.__yref_0 = yref_0

    @W_0.setter
    def W_0(self, W_0):
        check_if_2d_nparray(W_0, "W_0")
        self.__W_0 = W_0

    @Vx_0.setter
    def Vx_0(self, Vx_0):
        check_if_2d_nparray(Vx_0, "Vx_0")
        self.__Vx_0 = Vx_0

    @Vu_0.setter
    def Vu_0(self, Vu_0):
        check_if_2d_nparray(Vu_0, "Vu_0")
        self.__Vu_0 = Vu_0

    @Vz_0.setter
    def Vz_0(self, Vz_0):
        check_if_2d_nparray(Vz_0, "Vz_0")
        self.__Vz_0 = Vz_0

    @cost_ext_fun_type_0.setter
    def cost_ext_fun_type_0(self, cost_ext_fun_type_0):
        if cost_ext_fun_type_0 in ['casadi', 'generic']:
            self.__cost_ext_fun_type_0 = cost_ext_fun_type_0
        else:
            raise Exception('Invalid cost_ext_fun_type_0 value, expected numpy array.')

    # Lagrange term
    @property
    def cost_type(self):
        """
        Cost type at intermediate shooting nodes (1 to N-1)
        -- string in {EXTERNAL, LINEAR_LS, NONLINEAR_LS, CONVEX_OVER_NONLINEAR}.
        Default: 'LINEAR_LS'.
        """
        return self.__cost_type

    @property
    def W(self):
        """:math:`W` - weight matrix at intermediate shooting nodes (1 to N-1).
        Default: :code:`np.zeros((0,0))`.
        """
        return self.__W

    @property
    def Vx(self):
        """:math:`V_x` - x matrix coefficient at intermediate shooting nodes (1 to N-1).
        Default: :code:`np.zeros((0,0))`.
        """
        return self.__Vx

    @property
    def Vu(self):
        """:math:`V_u` - u matrix coefficient at intermediate shooting nodes (1 to N-1).
        Default: :code:`np.zeros((0,0))`.
        """
        return self.__Vu

    @property
    def Vz(self):
        """:math:`V_z` - z matrix coefficient at intermediate shooting nodes (1 to N-1).
        Default: :code:`np.zeros((0,0))`.
        """
        return self.__Vz

    @property
    def yref(self):
        r""":math:`y_\text{ref}` - reference at intermediate shooting nodes (1 to N-1).
        Default: :code:`np.array([])`.
        """
        return self.__yref

    @property
    def Zl(self):
        """:math:`Z_l` - diagonal of Hessian wrt lower slack at intermediate shooting nodes (0 to N-1).
        Default: :code:`np.array([])`.
        """
        return self.__Zl

    @property
    def Zu(self):
        """:math:`Z_u` - diagonal of Hessian wrt upper slack at intermediate shooting nodes (0 to N-1).
        Default: :code:`np.array([])`.
        """
        return self.__Zu

    @property
    def zl(self):
        """:math:`z_l` - gradient wrt lower slack at intermediate shooting nodes (0 to N-1).
        Default: :code:`np.array([])`.
        """
        return self.__zl

    @property
    def zu(self):
        """:math:`z_u` - gradient wrt upper slack at intermediate shooting nodes (0 to N-1).
        Default: :code:`np.array([])`.
        """
        return self.__zu

    @property
    def cost_ext_fun_type(self):
        """Type of external function for cost at intermediate shooting nodes (1 to N-1).
        -- string in {casadi, generic}
        Default: :code:'casadi'.
        """
        return self.__cost_ext_fun_type

    @cost_type.setter
    def cost_type(self, cost_type):
        cost_types = ('LINEAR_LS', 'NONLINEAR_LS', 'EXTERNAL', 'CONVEX_OVER_NONLINEAR')
        if cost_type in cost_types:
            self.__cost_type = cost_type
        else:
            raise Exception('Invalid cost_type value.')

    @cost_type_0.setter
    def cost_type_0(self, cost_type_0):
        cost_types = ('LINEAR_LS', 'NONLINEAR_LS', 'EXTERNAL', 'CONVEX_OVER_NONLINEAR')
        if cost_type_0 in cost_types:
            self.__cost_type_0 = cost_type_0
        else:
            raise Exception('Invalid cost_type_0 value.')

    @W.setter
    def W(self, W):
        check_if_2d_nparray(W, "W")
        self.__W = W


    @Vx.setter
    def Vx(self, Vx):
        check_if_2d_nparray(Vx, "Vx")
        self.__Vx = Vx

    @Vu.setter
    def Vu(self, Vu):
        check_if_2d_nparray(Vu, "Vu")
        self.__Vu = Vu

    @Vz.setter
    def Vz(self, Vz):
        check_if_2d_nparray(Vz, "Vz")
        self.__Vz = Vz

    @yref.setter
    def yref(self, yref):
        yref = check_if_nparray_and_flatten(yref, "yref")
        self.__yref = yref

    @Zl.setter
    def Zl(self, Zl):
        if isinstance(Zl, np.ndarray):
            self.__Zl = Zl
        else:
            raise Exception('Invalid Zl value, expected numpy array.')

    @Zu.setter
    def Zu(self, Zu):
        if isinstance(Zu, np.ndarray):
            self.__Zu = Zu
        else:
            raise Exception('Invalid Zu value, expected numpy array.')

    @zl.setter
    def zl(self, zl):
        if isinstance(zl, np.ndarray):
            self.__zl = zl
        else:
            raise Exception('Invalid zl value, expected numpy array.')

    @zu.setter
    def zu(self, zu):
        if isinstance(zu, np.ndarray):
            self.__zu = zu
        else:
            raise Exception('Invalid zu value, expected numpy array.')

    @cost_ext_fun_type.setter
    def cost_ext_fun_type(self, cost_ext_fun_type):
        if cost_ext_fun_type in ['casadi', 'generic']:
            self.__cost_ext_fun_type = cost_ext_fun_type
        else:
            raise Exception("Invalid cost_ext_fun_type value, expected one in ['casadi', 'generic'].")

    # Mayer term
    @property
    def cost_type_e(self):
        """
        Cost type at terminal shooting node (N)
        -- string in {EXTERNAL, LINEAR_LS, NONLINEAR_LS, CONVEX_OVER_NONLINEAR}.
        Default: 'LINEAR_LS'.
        """
        return self.__cost_type_e

    @property
    def W_e(self):
        """:math:`W_e` - weight matrix at terminal shooting node (N).
        Default: :code:`np.zeros((0,0))`.
        """
        return self.__W_e

    @property
    def Vx_e(self):
        """:math:`V_x^e` - x matrix coefficient for cost at terminal shooting node (N).
        Default: :code:`np.zeros((0,0))`.
        """
        return self.__Vx_e

    @property
    def yref_e(self):
        r""":math:`y_\text{ref}^e` - cost reference at terminal shooting node (N).
        Default: :code:`np.array([])`.
        """
        return self.__yref_e

    @property
    def Zl_e(self):
        """:math:`Z_l^e` - diagonal of Hessian wrt lower slack at terminal shooting node (N).
        Default: :code:`np.array([])`.
        """
        return self.__Zl_e

    @property
    def Zu_e(self):
        """:math:`Z_u^e` - diagonal of Hessian wrt upper slack at terminal shooting node (N).
        Default: :code:`np.array([])`.
        """
        return self.__Zu_e

    @property
    def zl_e(self):
        """:math:`z_l^e` - gradient wrt lower slack at terminal shooting node (N).
        Default: :code:`np.array([])`.
        """
        return self.__zl_e

    @property
    def zu_e(self):
        """:math:`z_u^e` - gradient wrt upper slack at terminal shooting node (N).
        Default: :code:`np.array([])`.
        """
        return self.__zu_e


    @property
    def Zl_0(self):
        """:math:`Z_l^0` - diagonal of Hessian wrt lower slack at initial node 0.
        Default: :code:`np.array([])`.
        """
        return self.__Zl_0

    @property
    def Zu_0(self):
        """:math:`Z_u^0` - diagonal of Hessian wrt upper slack at initial node 0.
        Default: :code:`np.array([])`.
        """
        return self.__Zu_0

    @property
    def zl_0(self):
        """:math:`z_l^0` - gradient wrt lower slack at initial node 0.
        Default: :code:`np.array([])`.
        """
        return self.__zl_0

    @property
    def zu_0(self):
        """:math:`z_u^0` - gradient wrt upper slack at initial node 0.
        Default: :code:`np.array([])`.
        """
        return self.__zu_0

    @property
    def cost_ext_fun_type_e(self):
        """Type of external function for cost at terminal shooting node (N).
        -- string in {casadi, generic}
        Default: :code:'casadi'.
        """
        return self.__cost_ext_fun_type_e

    @cost_type_e.setter
    def cost_type_e(self, cost_type_e):
        cost_types = ('LINEAR_LS', 'NONLINEAR_LS', 'EXTERNAL', 'CONVEX_OVER_NONLINEAR')
        if cost_type_e in cost_types:
            self.__cost_type_e = cost_type_e
        else:
            raise Exception('Invalid cost_type_e value.')

    @W_e.setter
    def W_e(self, W_e):
        check_if_2d_nparray(W_e, "W_e")
        self.__W_e = W_e

    @Vx_e.setter
    def Vx_e(self, Vx_e):
        check_if_2d_nparray(Vx_e, "Vx_e")
        self.__Vx_e = Vx_e

    @yref_e.setter
    def yref_e(self, yref_e):
        yref_e = check_if_nparray_and_flatten(yref_e, "yref_e")
        self.__yref_e = yref_e

    @Zl_e.setter
    def Zl_e(self, Zl_e):
        Zl_e = check_if_nparray_and_flatten(Zl_e, "Zl_e")
        self.__Zl_e = Zl_e

    @Zu_e.setter
    def Zu_e(self, Zu_e):
        Zu_e = check_if_nparray_and_flatten(Zu_e, "Zu_e")
        self.__Zu_e = Zu_e

    @zl_e.setter
    def zl_e(self, zl_e):
        zl_e = check_if_nparray_and_flatten(zl_e, "zl_e")
        self.__zl_e = zl_e

    @zu_e.setter
    def zu_e(self, zu_e):
        zu_e = check_if_nparray_and_flatten(zu_e, "zu_e")
        self.__zu_e = zu_e

    @Zl_0.setter
    def Zl_0(self, Zl_0):
        Zl_0 = check_if_nparray_and_flatten(Zl_0, "Zl_0")
        self.__Zl_0 = Zl_0

    @Zu_0.setter
    def Zu_0(self, Zu_0):
        Zu_0 = check_if_nparray_and_flatten(Zu_0, "Zu_0")
        self.__Zu_0 = Zu_0

    @zl_0.setter
    def zl_0(self, zl_0):
        zl_0 = check_if_nparray_and_flatten(zl_0, "zl_0")
        self.__zl_0 = zl_0

    @zu_0.setter
    def zu_0(self, zu_0):
        zu_0 = check_if_nparray_and_flatten(zu_0, "zu_0")
        self.__zu_0 = zu_0

    @cost_ext_fun_type_e.setter
    def cost_ext_fun_type_e(self, cost_ext_fun_type_e):
        if cost_ext_fun_type_e in ['casadi', 'generic']:
            self.__cost_ext_fun_type_e = cost_ext_fun_type_e
        else:
            raise Exception("Invalid cost_ext_fun_type_e value, expected one in ['casadi', 'generic'].")

    def set(self, attr, value):
        setattr(self, attr, value)



# This file is part of OMG-tools.
#
# OMG-tools -- Optimal Motion Generation-tools
# Copyright (C) 2016 Ruben Van Parys & Tim Mercy, KU Leuven.
# All rights reserved.
#
# OMG-tools is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import functools
#import cvxopt
import numpy as np
import scipy.linalg as la
import casadi as cas
from scipy.sparse import csr_matrix
# from piecewise import PiecewisePolynomial as ppoly
# from scipy.sparse.linalg import spsolve
from collections import Counter
import hashlib

def md5(data):
  m = hashlib.md5()
  m.update(data)
  return m.digest()

NO_POINTS = 501


def memoize(f):
    """ Memoization decorator"""
    class memodict(dict):
        def __init__(self, f):
            self.f = f

        def __call__(self, *args):
            key = (args[0], md5(np.atleast_1d(args[1])))
            if key in self:
                return self[key]
            else:
                ret = self[key] = self.f(*args)
                return ret

        def __get__(self, obj, objtype):
            return functools.partial(self.__call__, obj)
    return memodict(f)


def cached_class(klass):
    """Decorator to cache class instances by constructor arguments.
    """
    cache = {}

    @functools.wraps(klass, assigned=('__name__', '__module__'), updated=())
    class _decorated(klass):
        __doc__ = klass.__doc__

        def __new__(cls, *args, **kwds):
            key = (cls,) + tuple([md5(np.atleast_1d(k)) for k in args]) + tuple(kwds.items())
            try:
                inst = cache.get(key, None)
            except TypeError:  # Can't cache this set of arguments
                inst = key = None
            if inst is None:
                inst = klass(*args, **kwds)
                inst.__class__ = cls
                if key is not None:
                    cache[key] = inst
            return inst

        def __init__(self, *args, **kwds):
            pass

    return _decorated


def get_module(var):
    """Return the module of the variable"""
    return getattr(type(var), '__module__', '').split('.')[0]


class csr_matrix_alt(csr_matrix):
    """Subclass csr_matrix to overload dot operator for MX/SX classes and
    cvxpy classes"""
    def __init__(self, *args, **kwargs):
        csr_matrix.__init__(self, *args, **kwargs)

    def dot(self, other):
        if isinstance(other, (cas.MX, cas.SX)):
            # compatible with casadi 3.0 -- added by ruben
            return cas.mtimes(cas.DM(csr_matrix(self)), other)
            # NOT COMPATIBLE WITH CASADI 2.4
            # return cas.DMatrix(csr_matrix(self)).mul(other)
        elif get_module(other) in ['cvxpy', 'cvxopt']:
            return cvxopt.sparse(cvxopt.matrix(self.toarray())) * other
            # A = self.tocoo()
            # B = cvxopt.spmatrix(
            #     A.data, A.row.tolist(), A.col.tolist(), A.shape
            #     )
            # return B * other
        else:
            try:  # Scipy sparse matrix
                return super(csr_matrix_alt, self).dot(other)
            except:  # Regular numpy matrix
                return np.dot(self.toarray(), other)


class Basis(object):
    """A generic spline basis with a knot sequence and degree
    """
    def __init__(self, knots, degree):
        self.knots = np.array(knots)
        self.degree = degree
        self._x = np.linspace(knots[0], knots[-1], NO_POINTS)

    def __len__(self):
        return len(self.knots) - self.degree - 1

    def __call__(self, x):
        return self.eval_basis(x)

    def _ind(self, i, x):
        """Indicator function between knots[i] and knots[i + 1]
        """
        if i < self.degree + 1 and self.knots[0] == self.knots[i]:
            return (x >= self.knots[i]) * (x <= self.knots[i + 1])
        return (x > self.knots[i]) * (x <= self.knots[i + 1])

    def _combine(self, other, degree):
        """Combine two bases to a new basis of specified degree"""
        c_self = Counter(self.knots)
        c_other = Counter(other.knots)
        breaks = set(self.knots).union(other.knots)
        # Should be corrected!
        multiplicity = [max(c_self.get(b, -np.inf) + degree - self.degree,
                            c_other.get(b, -np.inf) + degree - other.degree)
                        for b in breaks]
        knots = sum([[b] * m for b, m in zip(breaks, multiplicity)], [])
        return self.__class__(sorted(knots), degree)

    def __add__(self, other):
        if isinstance(other, self.__class__):
            degree = max(self.degree, other.degree)
            return self._combine(other, degree)
        elif isinstance(other, float) or isinstance(other, int):
            return self
        else:
            raise TypeError("Not a basis error")

    __radd__ = __add__
    __sub__ = __add__
    __rsub__ = __sub__

    def __mul__(self, other):
        if isinstance(other, self.__class__):
            degree = self.degree + other.degree
            return self._combine(other, degree)
        elif isinstance(other, float) or isinstance(other, int):
            return self
        else:
            raise TypeError("Not a basis error")

    __rmul__ = __mul__

    def __pow__(self, pow):
        if isinstance(pow, int):
            degree = pow * self.degree
            return self._combine(self, degree)
        else:
            raise TypeError("Power must be integer")

    def __eq__(self, other):
        return self.knots.shape == other.knots.shape and all(self.knots == other.knots) and self.degree == other.degree

    def __hash__(self):
        sh = [hash(self.degree)] + [hash(e) for e in self.knots]
        r = sh[0]
        for e in sh[1:]:
          r = r ^ e
        return r

    def insert_knots(self, knots):
        unique_knots = np.setdiff1d(knots, self.knots)
        knots = np.sort(np.append(self.knots, unique_knots))
        return self.__class__(knots, self.degree)

    def greville(self):
        """Return the Greville abscissae of the basis"""
        return [1. / self.degree * sum(self.knots[k + 1:k + self.degree + 1])
                for k in range(len(self))]

    def scale(self, factor, shift=0):
        # by default the domain is [0,1]
        # this function scales the basis domain and shifts it
        knots = self.knots*factor + shift
        return self.__class__(knots, self.degree)


@cached_class
class BSplineBasis(Basis):
    """
    A numerical Bspline basis
    """
    @memoize
    def eval_basis(self, x):
        """Evaluate the BSplineBasis at x.

        This function implements the Cox-de Boor formula for B-splines
        """
        x = np.array(x)
        k = self.knots
        basis = [[self._ind(i, x) * 1.0 for i in range(len(k) - 1)]]
        for d in range(1, self.degree + 1):
            basis.append([])
            for i in range(len(k) - d - 1):
                b = 0 * x
                bottom = k[i + d] - k[i]
                if bottom != 0:
                    b = (x - k[i]) * basis[d - 1][i] / bottom
                bottom = k[i + d + 1] - k[i + 1]
                if bottom != 0:
                    b += (k[i + d + 1] - x) * basis[d - 1][i + 1] / bottom
                basis[-1].append(b)
        # Consider sparse matrices?
        return csr_matrix_alt(np.c_[basis[-1]].T)

    def derivative(self, o=1):
        """Returns derivative of the basisfunctions

        Computes the derivative using eq. (16) in [de Boor, Chapter X, 2001].

        Args:
            x (numpy.array): grid on which to evaluate basisfunctions
            o (int): order of the derivative (default is 1)

        Returns:
            Numpy.array: columns contain the value of the derivative of the
                basisfunction evaluated at x
        """
        B = self.__class__(self.knots[o:-o], self.degree - o)
        P = np.eye(len(self))
        knots = self.knots
        for i in range(o):
            knots = knots[1:-1]
            delta_knots = knots[self.degree - i:] - knots[:- self.degree + i]
            T = np.zeros((len(self) - 1 - i, len(self) - i))
            j = np.arange(len(self) - 1 - i)
            T[(j, j)] = -1. / delta_knots
            T[(j, j + 1)] = 1. / delta_knots
            P = (self.degree - i) * np.dot(T, P)
        return B, csr_matrix_alt(P)

    def support(self):
        """Return a list of support intervals for each basis function"""
        return list(zip(
            self.knots[:-(self.degree + 1)],
            self.knots[(self.degree + 1):]
            ))

    def pairs(self, other):
        """Return which pairs remain when multiplying two bases"""
        def is_valid(a, b):
            """Return True if intervals a, b overlap"""
            return max(a[0], b[0]) < min(a[1], b[1])
        i_self = self.support()
        i_other = other.support()
        pairs = np.where([[is_valid(j, x) for x in i_other]
                          for j in i_self])
        # Additionaly build a selection matrix for the product
        S = np.zeros((len(self), len(self) * len(other)))
        # S[[pairs[0], pairs[0] * len(self) + pairs[1]]] = 1.
        return pairs, S

    def transform(self, other, TOL=1e-10):
        """Transformation from one basis to another.

        Returns a transformation matrix T such that

            self(x).T = other(x)

        TODO: Can we use the greville points instead of max?
        """
        b = self(self._x).toarray()
        m = np.argmax(b, axis=0)
        # x = np.linspace(self.knots[0], self.knots[-1], NO_POINTS)
        xmax = self._x[m]
        if isinstance(other, BSplineBasis):
            # if (self.knots[0] == other.knots[0] and
            #    self.knots[-1] == other.knots[-1]):
            T = la.solve(b[m, :], other(xmax).toarray())
        else:
            try:
                T = la.solve(b[m, :], other(xmax))
            except:  # In case of multiplication
                T = la.solve(b[m, :], other(m))
        T[abs(T) < TOL] = 0.
        return csr_matrix_alt(T)

    def as_poly(self):
        """Returns polynomial description of the basis functions"""
        k = self.knots
        k_min, k_max = min(self.knots), max(self.knots)
        basis = [[ppoly([a, b], [[1]]) for (a, b) in zip(k[:-1], k[1:])]]
        for d in range(1, self.degree + 1):
            basis.append([])
            for i in range(len(k) - d - 1):
                b = ppoly([k_min, k_max], [0])
                bottom = k[i + d] - k[i]
                if bottom != 0:
                    b += ppoly([k_min, k_max], [[-k[i], 1]]) * basis[d - 1][i] * (1. / bottom)
                bottom = k[i + d + 1] - k[i + 1]
                if bottom != 0:
                    b += ppoly([k_min, k_max], [[k[i + d + 1], -1]]) * basis[d - 1][i + 1] * (1. / bottom)
                basis[-1].append(b)
        return basis[-1]


class NurbsBasis(Basis):
    def __init__(self, knots, degree, weights):
        self.weights = weights
        self.bbasis = BSplineBasis(knots, degree)
        super(NurbsBasis, self).__init__(knots, degree)

    def eval_basis(self, x):
        B = self.bbasis(x)
        denom = B.dot(self.weights)
        if isinstance(self.weights, cas.MX):
            pass
            # B.dot(cas.diag(self.weights))
        else:
            return ((B.toarray() * self.weights).T / denom).T


class TSplineBasis(Basis):
    """A trigonometric spline basis"""
    def eval_basis(self, x):
        """
        Basisfunction of degree d evaluated on x
        """
        k = self.knots
        basis = [[self._ind(i, x) * 1.0 for i in range(len(k) - 1)]]
        for d in range(1, self.degree + 1):
            basis.append([])
            for i in range(len(k) - d - 1):
                b = 0 * x
                bottom = np.sin(0.5 * (k[i + d] - k[i]))
                if bottom != 0:
                    b = np.sin(0.5 * (x - k[i])) * basis[d - 1][i] / bottom
                bottom = np.sin(0.5 * (k[i + d + 1] - k[i + 1]))
                if bottom != 0:
                    b += np.sin(0.5 * (k[i + d + 1] - x)) * basis[d - 1][i + 1] / bottom
                basis[-1].append(b)
        return csr_matrix_alt(np.c_[basis[-1]].T)


class Spline(object):
    def __init__(self, basis, coeffs):
        # self.coeffs = np.array(coeffs).ravel()
        self.coeffs = coeffs
        self.basis = basis
        # if isinstance(coeffs, (cas.SXMatrix, cas.SX)):
        #     self.basis._basis = cas.DMatrix(self.basis._basis)

    def __call__(self, x):
        return self.basis(x).dot(self.coeffs)

    def __len__(self):
        return len(self.basis)

    def __eq__(self, other):
        return (self.basis == other.basis and
                type(self.coeffs) == type(other.coeffs) and
                all(self.coeffs == other.coeffs))

    def __hash__(self):
        sh = [hash(self.basis)] + [hash(e) for e in self.coeffs]
        r = sh[0]
        for e in sh[1:]:
          r = r ^ e
        return r


class BSpline(Spline):
    """Construct a Bspline curve from the basis B and coefficients c
    """
    def __add__(self, other):
        return self.common(other, lambda a,b : a+b)

    def common(self, other, op):
        if isinstance(other, self.__class__):
            basis = self.basis + other.basis
            coeffs = op(basis.transform(self.basis).dot(self.coeffs),
                      basis.transform(other.basis).dot(other.coeffs))
        else:
            try:
                basis = self.basis
                coeffs = op(self.coeffs,other)  # Only for BSpline!
            except:
                NotImplementedError("Incompatible datatype")
        return self.__class__(basis, coeffs)

    __radd__ = __add__

    def __neg__(self):
        return self.__class__(self.basis, -self.coeffs)

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return other + (-self)

    def __mul__(self, other):
        if isinstance(other, self.__class__):
            basis = self.basis * other.basis
            pairs, S = self.basis.pairs(other.basis)
            b_self = self.basis(basis._x)
            b_other = other.basis(basis._x)
            basis_product = b_self[:, pairs[0]].multiply(b_other[:, pairs[1]])
            T = basis.transform(lambda y: basis_product.toarray()[y, :])
            try:
                coeffs_product = (self.coeffs[pairs[0].tolist()] *
                                  other.coeffs[pairs[1].tolist()])
            except:  # cvxopt, cvxpy, assuming other.coeffs is not a variable
                S = np.zeros((len(pairs[0]), len(self)))
                S[[list(range(len(pairs[0]))), pairs[0]]] = 1.
                S = cvxopt.matrix(S)
                coeffs_product = cvxopt.spdiag(other.coeffs[pairs[1].tolist()]) * S * self.coeffs
                # coeffs_product = cp.vstack(*[self.coeffs[p0] * other.coeffs[p1] for (p0, p1) in zip(*pairs)])
            return self.__class__(basis, T.dot(coeffs_product))
        else:
            try:
                basis = self.basis
                coeffs = other * self.coeffs
                return self.__class__(basis, coeffs)
            except:
                NotImplementedError("Incompatible datatype")

    def __rmul__(self, other):
        return self.__mul__(other)

    def __pow__(self, power):
        """A naive implementation of the power function..."""
        if isinstance(power, int):
            a = self
            for i in range(1, power):
                a *= self
            return a
        else:
            TypeError("Exponent must be integer")

    def __div__(self, other):
        basis = self.basis + other.basis
        weights = basis.transform(other.basis).dot(other.coeffs)
        coeffs = basis.transform(self.basis).dot(self.coeffs) / weights
        return Nurbs(NurbsBasis(basis.knots, basis.degree, weights), coeffs)

    def derivative(self, o=1):
        if o == 0:
            return self
        else:
            Bd, Pd = self.basis.derivative(o=o)
            return self.__class__(Bd, Pd.dot(self.coeffs))

    def insert_knots(self, knots):
        """Returns an equivalent spline with knot insertion"""
        basis = self.basis.insert_knots(knots)
        coeffs = basis.transform(self.basis).dot(self.coeffs)
        return self.__class__(basis, coeffs)

    def integral(self):
        """Returns the value of the integral over the support.

        This is a literal implementation of formula X.33 from deBoor and
        assumes that at x = knots[-1], only the last basis function is active
        """
        knots = self.basis.knots
        coeffs = self.coeffs
        d = self.basis.degree
        K = csr_matrix_alt(np.diag((knots[d + 1:] - knots[:-(d + 1)]) / (d + 1)))
        try:
            return sum(K.dot(coeffs))
        except:
            return cas.sum1(K.dot(coeffs))
        # try:
        #     return sum(coeffs * (knots[d + 1:] - knots[:-(d + 1)])) / (d + 1)

    def roots(self):
        """Return the roots of the B-spline

        Algorithm:
        * Determine polynomial description
        * Determine roots of each polynomial subpiece
        * Check if root is in the support of the polynomial piece
        """
        basis = self.basis.as_poly()
        spline = np.sum([self.coeffs[i] * basis[i] for i in range(len(basis))])
        roots = []
        for i, f in enumerate(spline.functions):
            root = f.roots()
            roots.extend([r for r in root
                         if spline.knots[i] <= r < spline.knots[i + 1]])
        return roots

    def scale(self, factor, shift=0):
        # by default the domain is [0,1]
        # this function scales the domain of the spline and shifts it
        basis = self.basis.scale(factor, shift=shift)
        return self.__class__(basis, self.coeffs)

    def __ge__(self, other):
        return self.common(other, lambda a,b : a>=b).coeffs

    def __gt__(self, other):
        return self.common(other, lambda a,b : a>b).coeffs

    def __le__(self, other):
        return self.common(other, lambda a,b : a<=b).coeffs

    def __lt__(self, other):
        return self.common(other, lambda a,b : a<b).coeffs

class Nurbs(Spline):
    def __init__(self, basis, coeffs):
        super(Nurbs, self).__init__(basis, coeffs)
        self.num = BSpline(self.basis.bbasis, self.coeffs * self.basis.weights)
        self.denom = BSpline(self.basis.bbasis, self.basis.weights)

    def __mul__(self, other):
        num = self.num * other
        return num / self.denom

    __rmul__ = __mul__

    def __add__(self, other):
        if isinstance(other, Nurbs):
            return (self.num * other.denom + self.denom * other.num) / (self.denom * other.denom)

    def derivative(self, o=1):
        """Derivative of a Nurbs"""
        if o == 1:
            b = self.basis.bbasis
            db = b.derivative()[0]
            # The denominator
            denom2 = self.denom ** 2
            # coeffs of the numerator
            pairs, S = b.pairs(db)
            dnum = self.num.derivative()
            ddenom = self.denom.derivative()
            coeffs_product = dnum.coeffs[pairs[1].tolist()] * self.denom.coeffs[pairs[0].tolist()] - self.num.coeffs[pairs[0].tolist()] * ddenom.coeffs[pairs[1].tolist()]
            bx = b(denom2.basis._x)
            dbx = db(denom2.basis._x)
            basis_product = bx[:, pairs[0]].multiply(dbx[:, pairs[1]])
            T = denom2.basis.transform(lambda y: basis_product.toarray()[y, :])
            coeffs = T.dot(coeffs_product) / denom2.coeffs
            basis = NurbsBasis(denom2.basis.knots, denom2.basis.degree, denom2.coeffs)
            return self.__class__(basis, coeffs)
            # num = BSpline(self.basis.bbasis, self.coeffs * self.basis.weights)
            # denom = BSpline(self.basis.bbasis, self.basis.weights)
            # Compute numerator efficiently
            # return ((dnum * denom - num * ddenom) /  # This numerator has twice the same basis!
            #         denom ** 2)  # Can we simplify this, make it faster? -> The same basis is created multiple times!
        else:
            return self.derivative().derivative(o=o-1)

    def insert_knots(self, knots):
        """Returns an equivalent spline with knot insertion"""
        b = self.basis.bbasis.insert_knots(knots)
        weights = b.transform(self.basis.bbasis).dot(self.basis.weights)
        coeffs = b.transform(self.basis.bbasis).dot(self.coeffs)
        basis = NurbsBasis(b.knots, b.degree, weights)
        return self.__class__(basis, coeffs)


class TensorBSpline(object):
    """A multidimensional spline"""
    def __init__(self, basis, coeffs, var):
        self.basis = tuple(basis)
        self.var = tuple(var)
        self.coeffs = coeffs

    # def _reduce(self):
    #     """Set irrelevant coefficients to zero"""
    #     def is_valid(a, b):
    #         """Return True if intervals a, b overlap"""
    #         return max(a[0], b[0]) < min(a[1], b[1])
    #     i = [zip(b.knots[:-(b.degree + 1)], b.knots[b.degree + 1:])
    #          for b in self.basis]
    #     map(lambda x: is_valid(j, x), ii) for j in i[0]
    #     pairs = np.where([map(lambda x: is_valid(j, x), i_other)
    #                       for j in i_self])

    def dims(self):
        """The number of dimensions of the spline"""
        return len(self.basis)

    def __call__(self, x):
        """Evaluate TensorBSpline
        There still seems to be something wrong here...
        """
        s = np.inner(self.basis[-1](x[-1]).toarray(), self.coeffs)
        for i in reversed(list(range(self.dims() - 1))):
            s = np.inner(self.basis[i](x[i]).toarray(), s)
        return s

    def __add__(self, other):
        if isinstance(other, TensorBSpline) and other.var == self.var:
            if self.dims() == 2 and get_module(self.coeffs) in ['cvxpy', 'cvxopt']:
                basis = list(map(lambda x, y: x + y, self.basis, other.basis))
                Tself = list(map(lambda x, y: cvxopt.matrix(x.transform(y).toarray()), basis, self.basis))
                Tother = list(map(lambda x, y: cvxopt.matrix(x.transform(y).toarray()), basis, other.basis))
                cself = Tself[0] * self.coeffs * Tself[1].T
                cother = Tother[0] * other.coeffs * Tother[1].T
                coeffs = cself + cother
            else:
                basis = list(map(lambda x, y: x + y, self.basis, other.basis))
                Tself = list(map(lambda x, y: x.transform(y).toarray(), basis, self.basis))
                Tother = list(map(lambda x, y: x.transform(y).toarray(), basis, other.basis))
                cself = self.coeffs
                for i in range(self.dims()):
                    cself = np.tensordot(Tself[i], cself.swapaxes(0, i), axes=[1, 0]).swapaxes(0, i)
                cother = other.coeffs
                for i in range(other.dims()):
                    cother = np.tensordot(Tother[i], cother.swapaxes(0, i), axes=[1, 0]).swapaxes(0, i)
                coeffs = cself + cother
        else:
            try:
                basis = self.basis
                coeffs = self.coeffs + other  # Only for BSpline!
            except:
                NotImplementedError("Incompatible datatype")
        return self.__class__(basis, coeffs, self.var)

    __radd__ = __add__

    def __neg__(self):
        return self.__class__(self.basis, -self.coeffs, self.var)

    def __sub__(self, other):
        return self + (-other)

    __rsub__ = __sub__

    def __mul__(self, other):
        if isinstance(other, TensorBSpline):
            if len(self.basis) > 2:
                return NotImplementedError("Too complex to implement :-)")
            basis = list(map(lambda x, y: x * y, self.basis, other.basis))
            pairs = list(map(lambda x, y: x.pairs(y)[0], self.basis, other.basis))
            basis_product = list(map(lambda x, y, p, b: x(b._x)[:, p[0]].multiply(y(b._x)[:, p[1]]), self.basis, other.basis, pairs, basis))
            coeffs_product = (self.coeffs[pairs[0][0]].T[pairs[1][0]] *
                              other.coeffs[pairs[0][1]].T[pairs[1][1]])
            T = list(map(lambda x, b: b.transform(lambda y: x.toarray()[y, :]), basis_product, basis))
            coeffs = coeffs_product.T
            for i, t in enumerate(T):
                coeffs = np.tensordot(t.toarray(), coeffs.swapaxes(0, i), axes=[1, 0]).swapaxes(0, i)
            return self.__class__(basis, coeffs, self.var)
            return self.coeffs[pairs[0][0].tolist(), pairs[1][0].tolist()] * other.coeffs[pairs[0][1].tolist(), pairs[1][1].tolist()]
            # coeffs_product = np.kron(self.coeffs, other.coeffs)

        else:
            try:
                basis = self.basis
                coeffs = self.coeffs * other
            except:
                NotImplementedError("Incompatible datatype")
        return self.__class__(basis, coeffs, self.var)

    __rmul__ = __mul__

    def integral(self):
        """Returns the value of the integral over the support.

        This is a literal implementation of formula X.33 from deBoor and
        assumes that at x = knots[-1], only the last basis function is active
        """
        knots = [b.knots for b in self.basis]
        coeffs = self.coeffs
        deg = [b.degree for b in self.basis]
        K = [(k[d + 1:] - k[:-(d + 1)]) / (d + 1)
             for (k, d) in zip(knots, deg)]
        if self.dims() == 2: #get_module(self.coeffs) in ['cvxpy', 'cvxopt']:
            i = cvxopt.matrix(K[0]).T * self.coeffs * cvxopt.matrix(K[1])
            return i
        i = np.inner(K[-1], coeffs)
        for ki in K[:-1]:
            i = np.inner(ki, i)
        return i
